import py_misc_utils.alog as alog
import py_misc_utils.core_utils as pycu
import py_misc_utils.utils as pyu

from .utils import *


class ExternLogic:

  def __init__(self, modname,
               funcname=None,
               nargs=None,
               params=None,
               args=None,
               name_remap=None,
               filename=None):
    self.modname = modname
    self.funcname = funcname
    self.nargs = nargs or 0
    self.params = params or ()
    self.args = args or dict()
    self.name_remap = name_remap or dict()
    self.filename = filename or modname


class ExternModule:

  def __init__(self, name):
    self.name = name
    self._logic = dict()

  def add_logic(self, name, logic):
    self._logic[name] = logic

  def get_logic(self, name):
    xlogic = self._logic.get(name)
    if xlogic is None:
      fatal(f'Unable to find {name} logic within {self.name} module')

    return xlogic

  @classmethod
  def load(cls, path):
    mdata = pyu.load_config(path)

    xmod = cls(mdata['name'])

    name_remap = mdata.get('name_remap')

    for fname, fdata in mdata.get('functions', dict()).items():
      xlogic = pycu.obj_from_dict(ExternLogic, fdata)
      if not xlogic.name_remap and name_remap:
        xlogic.name_remap = name_remap

      xmod.add_logic(fname, xlogic)

    return xmod




import argparse
import os
import re
import shutil
import string
import subprocess
import tempfile
import textwrap

import py_misc_utils.alog as alog
import py_misc_utils.app_main as app_main
import py_misc_utils.utils as pyu


class Verifier(object):

  def __init__(self, binary, cmdline_args):
    xpath = shutil.which(binary)
    if not xpath:
      alog.debug(f'Unable to find binary "{binary}" for {self.name} verifier')
      raise NotImplementedError(f'Unable to find binary "{binary}" for {self.name} verifier')

    alog.info(f'Found {self.name} verifier at {xpath}')

    self._binary = binary
    self._xpath = xpath
    self._args = cmdline_args

  def _make_subs_ctx(self, files, backend, top_entity, **kwargs):
    sctx = {
      'CSFILES': ', '.join(files),
      'SFILES': ' '.join(files),
      'TOP': top_entity,
      'BACKEND': backend,
    }
    sctx.update(kwargs)

    return sctx


class VivadoVerifier(Verifier):

  BINARY = 'vivado'
  CMDLINE = '-mode batch -nolog -nojournal -source'

  def __init__(self, cmdline_args):
    super().__init__(self.BINARY, cmdline_args)
    self._backend_read = {
      'verilog': ['read_verilog -sv {{ $CSFILES }}'],
      'vhdl': ['read_vhdl -vhdl2008 {{ $CSFILES }}'],
    }

  def _create_script(self, files, backend):
    script = list(self._backend_read[backend])

    script.append('synth_design -top $TOP')

    return '\n'.join(script)

  @property
  def name(self):
    return 'Vivado'

  @property
  def backends(self):
    return ('verilog', 'vhdl')

  def verify(self, files, backend, top_entity):
    with tempfile.TemporaryDirectory() as tmp_path:
      script = self._create_script(files, backend)

      sctx = self._make_subs_ctx(files, backend, top_entity)

      script = string.Template(script).substitute(**sctx)

      alog.debug(f'Vivado Script:\n{textwrap.indent(script, "  ")}')

      fd, path = tempfile.mkstemp(dir=tmp_path, suffix='.tcl', text=True)
      with os.fdopen(fd, mode='wt') as tfd:
        tfd.write(script)

      cmdline = re.split(r'\s+', string.Template(self.CMDLINE).substitute(**sctx))

      try:
        output = subprocess.check_output([self._xpath] + cmdline + [path],
                                         stderr=subprocess.STDOUT)
      except subprocess.CalledProcessError as ex:
        pyu.fatal(f'Verification process exited with {ex.returncode} code. ' \
                  f'Error output:\n' + ex.output.decode())

      return output


class GhdlVerifier(Verifier):

  BINARY = 'ghdl'
  CMDLINE = '-a --std=08 --workdir=$WORKDIR -frelaxed -Wno-shared'

  def __init__(self, cmdline_args):
    super().__init__(self.BINARY, cmdline_args)

  @property
  def name(self):
    return 'GHDL'

  @property
  def backends(self):
    return ('vhdl',)

  def verify(self, files, backend, top_entity):
    with tempfile.TemporaryDirectory() as tmp_path:
      sctx = self._make_subs_ctx(files, backend, top_entity,
                                 WORKDIR=tmp_path)

      cmdline = re.split(r'\s+', string.Template(self.CMDLINE).substitute(**sctx))

      try:
        output = subprocess.check_output([self._xpath] + cmdline + list(files),
                                         stderr=subprocess.STDOUT)
      except subprocess.CalledProcessError as ex:
        pyu.fatal(f'Verification process exited with {ex.returncode} code. ' \
                  f'Error output:\n' + ex.output.decode())

      return output


class VerilatorVerifier(Verifier):

  BINARY = 'verilator'
  CMDLINE = '--lint-only --timing -sv --top $TOP'

  def __init__(self, cmdline_args):
    super().__init__(self.BINARY, cmdline_args)

  @property
  def name(self):
    return 'Verilator'

  @property
  def backends(self):
    return ('verilog',)

  def verify(self, files, backend, top_entity):
    with tempfile.TemporaryDirectory() as tmp_path:
      sctx = self._make_subs_ctx(files, backend, top_entity)

      cmdline = re.split(r'\s+', string.Template(self.CMDLINE).substitute(**sctx))

      try:
        output = subprocess.check_output([self._xpath] + cmdline + list(files),
                                         stderr=subprocess.STDOUT)
      except subprocess.CalledProcessError as ex:
        pyu.fatal(f'Verification process exited with {ex.returncode} code. ' \
                  f'Error output:\n' + ex.output.decode())

      return output


class SlangVerifier(Verifier):

  BINARY = 'slang'
  CMDLINE = '-q --std 1800-2017 --top $TOP'

  def __init__(self, cmdline_args):
    super().__init__(self.BINARY, cmdline_args)

  @property
  def name(self):
    return 'Slang'

  @property
  def backends(self):
    return ('verilog',)

  def verify(self, files, backend, top_entity):
    with tempfile.TemporaryDirectory() as tmp_path:
      sctx = self._make_subs_ctx(files, backend, top_entity)

      cmdline = re.split(r'\s+', string.Template(self.CMDLINE).substitute(**sctx))

      try:
        output = subprocess.check_output([self._xpath] + cmdline + list(files),
                                         stderr=subprocess.STDOUT)
      except subprocess.CalledProcessError as ex:
        pyu.fatal(f'Verification process exited with {ex.returncode} code. ' \
                  f'Error output:\n' + ex.output.decode())

      return output


class YosysVerifier(Verifier):

  BINARY = 'yosys'
  CMDLINE = '-q -s'

  def __init__(self, cmdline_args):
    super().__init__(self.BINARY, cmdline_args)
    self._backends = ['verilog']
    self._plugins = []
    self._backend_read = {
      'verilog': ['read_verilog -sv $SFILES'],
    }
    self._find_plugins()

  def _check_plugin(self, name, search_paths):
    for path in search_paths:
      ppath = os.path.join(path, f'{name}.so')
      if os.path.exists(ppath):
        return ppath

  def _find_plugins(self):
    bin_path = os.path.dirname(self._xpath)
    search_paths = [
      bin_path,
      os.path.join(os.path.dirname(bin_path), 'lib'),
    ]
    if path := os.getenv('YOSYS_PLUGINS_PATH'):
      search_paths.append(os.path.abspath(path))

    alog.debug(f'Yosys Search Path: {search_paths}')

    if ppath := self._check_plugin('ghdl', search_paths):
      self._plugins.append(ppath)
      self._backends.append('vhdl')
      self._backend_read['vhdl'] = ['ghdl --std=08 $SFILES -e $TOP']

    if ppath := self._check_plugin('slang', search_paths):
      self._plugins.append(ppath)
      self._backend_read['verilog'] = [('read_slang --single-unit --allow-use-before-declare '
                                        '--allow-hierarchical-const --top $TOP $SFILES')]

  def _create_script(self, files, backend):
    script = list(self._backend_read[backend])

    script.append('hierarchy -check -top $TOP')
    script.append('check')
    script.append('synth')

    return '\n'.join(script)

  @property
  def name(self):
    return 'Yosys'

  @property
  def backends(self):
    return tuple(sorted(self._backends))

  def verify(self, files, backend, top_entity):
    with tempfile.TemporaryDirectory() as tmp_path:
      script = self._create_script(files, backend)

      sctx = self._make_subs_ctx(files, backend, top_entity)

      script = string.Template(script).substitute(**sctx)

      alog.debug(f'Yosys Script:\n{textwrap.indent(script, "  ")}')

      fd, path = tempfile.mkstemp(dir=tmp_path, suffix='.ys', text=True)
      with os.fdopen(fd, mode='wt') as tfd:
        tfd.write(script)

      cmdline = re.split(r'\s+', string.Template(self.CMDLINE).substitute(**sctx))

      plugins = []
      for mpath in self._plugins:
        plugins.extend(['-m', mpath])

      try:
        output = subprocess.check_output([self._xpath] + plugins + cmdline + [path],
                                         stderr=subprocess.STDOUT)
      except subprocess.CalledProcessError as ex:
        pyu.fatal(f'Verification process exited with {ex.returncode} code. ' \
                  f'Error output:\n' + ex.output.decode())

      return output


VERIFY_TOOLS = {
  # Ensure you have sourced the Vivado settings shell script which sets up the
  # proper environment variables (usually named settings64.sh).
  'Vivado': VivadoVerifier,
  'GHDL': GhdlVerifier,
  'Verilator': VerilatorVerifier,
  'Slang': SlangVerifier,
  'Yosys': YosysVerifier,
}

def _load_verifiers(args):
  verifiers, exclude = [], set(args.exclude or [])

  for name, vclass in VERIFY_TOOLS.items():
    if name.lower() not in exclude:
      try:
        verifier = vclass(args)

        if args.backend in verifier.backends:
          alog.info(f'Adding {verifier.name} verifier')
          verifiers.append(verifier)
      except NotImplementedError:
        pass

  return verifiers


def _main(args):
  verifiers = _load_verifiers(args)

  if not verifiers:
    pyu.fatal(f'Unable to find any valid HDL verification tools')

  for verifier in verifiers:
    alog.info(f'Running {verifier.name} verifier on {args.backend} files {args.inputs}')

    verifier.verify(args.inputs, args.backend.lower(), args.entity)


if __name__ == '__main__':
  parser = argparse.ArgumentParser(description='VHDL/Verilog Code Verifier',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--inputs', nargs='+', required=True,
                      help='The input files to be analyzed')
  parser.add_argument('--entity', type=str, required=True,
                      help='The root entity name')
  parser.add_argument('--backend', type=str, default='vhdl',
                      choices={'vhdl', 'verilog'},
                      help='The backend to generate the code for')
  parser.add_argument('--exclude', action='append',
                      choices=set(t.lower() for t in VERIFY_TOOLS.keys()),
                      help='The list of verifiers to be excluded')

  app_main.main(parser, _main)




from .common_defs import *
from .decorators import *
from .entity import *
from .interface import *
from .port import *
from .pyxhdl import *
from .types import *
from .utils import *
from .vars import *
from .verilog_emitter import *
from .vhdl_emitter import *
from .wrap import *



import collections


class _Phi:

  def __init__(self, parent):
    self._parent = parent
    self._phid = dict()

  def set_version(self, name, ver):
    cver = self._phid.get(name, 0)
    if cver >= ver:
      raise ValueError(f'New version must be greater: {cver} >= {ver}')

    self._phid[name] = ver

  def get_version(self, name):
    ver = self._phid.get(name, 0)
    # The root Phi does not generate new versions, only collects commits.
    if ver == 0 and self._parent is not None:
      pphi = self._parent
      while pphi is not None:
        ver = pphi._phid.get(name, 0)
        if ver > 0:
          break

        pphi = pphi._parent

      ver += 1
      self._phid[name] = ver

    return ver

  def cur_version(self, name):
    pphi = self
    while pphi is not None:
      ver = pphi._phid.get(name, 0)
      if ver > 0:
        return ver

      pphi = pphi._parent

    return 0

  def __len__(self):
    return len(self._phid)

  def __iter__(self):
    return iter(self._phid.items())


class Phis:

  def __init__(self):
    self._phis = [_Phi(None)]

  @staticmethod
  def versioned_name(name, ver):
    return f'{name}_{ver}' if ver else name

  @property
  def top_phi(self):
    return self._phis[-1]

  def push_phi(self):
    phi = _Phi(self.top_phi)
    self._phis.append(phi)

    return phi

  def pop_phi(self):
    return self._phis.pop()

  def get_version(self, name):
    return self.top_phi.get_version(name)

  def cur_version(self, name):
    return self.top_phi.cur_version(name)

  def commit(self, commits):
    cphi = self.top_phi
    for name, cver in commits.items():
      cphi.set_version(name, cver)

  def flush(self, branches_phis, assign_fn):
    commits = collections.defaultdict(int)
    for bphi in branches_phis:
      for name, ver in bphi:
        commits[name] = max(commits[name], ver)

    for bphi in branches_phis:
      for name, cver in commits.items():
        bver = bphi.cur_version(name)
        if cver > bver:
          # Assign FROM (bver), TO (cver) versions.
          assign_fn(bphi, name, bver, cver)

    self.commit(commits)

  def load_var(self, name):
    ver = self.cur_version(name)
    return Phis.versioned_name(name, ver)

  def store_var(self, name):
    ver = self.get_version(name)
    return Phis.versioned_name(name, ver)

  def __len__(self):
    return len(self._phis)




# def load_var(name):
#   vname = phis.load_var(name)
#   # Load vname
#   ...


# def handle_assign(name, value):
#   var = load_var(name)
#   if isinstance(var, Value):
#     vname = phis.store_var(name)
#     # Assign vname = value
#     ...


# def put_assign(place, name, from_ver, to_ver):
#   from_name = Phis.versioned_name(name, from_ver)
#   to_name = Phis.versioned_name(name, to_ver)
#   # Emit an assigment to_name = from_name
#   ...



# def handle_if(node):
#   emit_if_cond(node)

#   iphi = phis.push_phi()
#   icode = emit_if_body(node)
#   pif = emit_placement()
#   phis.pop_phi()

#   ephi = phis.push_phi()
#   ecode = emit_else_body(node)
#   pelse = emit_placement()
#   phis.pop_phi()

#   fdist = {id(iphi): pif, id(ephi): pelse}

#   def fassign(phi, name, from_ver, to_ver):
#     place = fdist[id(phi)]
#     put_assign(place, name, from_ver, to_ver)

#   phis.flush((iphi, ephi), fassign)




import re

import py_misc_utils.alog as alog
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .pyxhdl import *
from .types import *
from .utils import *
from .vars import *


def argn_dtype(n):
  def typefn(args, kwargs):
    return args[n].dtype

  return typefn


def create_function(fnname, fnmap, fnsig=None, dtype=None):
  return _ExternalFunction(fnname, fnmap, fnsig=fnsig, dtype=dtype)


class _Marshal:

  BITS_CLASSES = {Uint, Sint, Bits}

  def __init__(self, mstr, fnname, argref):
    self._fnname = fnname
    self._argref = argref
    self._tmatch = TypeMatcher.parse(mstr)

  def _get_tclass_dtype(self, tclass, dtype):
    if not has_bits(tclass):
      return tclass()
    if tclass in self.BITS_CLASSES and type(dtype) in self.BITS_CLASSES:
      return tclass(dtype.nbits)

  def _tclass_cast(self, ctx, arg, tclass):
    if isinstance(arg, Value):
      dtype = self._get_tclass_dtype(tclass, arg.dtype)
      if dtype is None:
        if not isinstance(arg.dtype, tclass):
          fatal(f'Wrong type for argument {self._argref} of {self._fnname}() ' \
                f'call: {pyiu.cname(tclass)} vs {pyiu.cname(arg.dtype)}')
      else:
        arg = ctx.emitter.cast(arg, dtype)
    else:
      arg = ctx.emitter.tclass_cast(tclass, arg)

    return arg

  def __call__(self, ctx, arg):
    return self._tmatch.cast(
      arg,
      dtype_fn=lambda v, t: ctx.emitter.cast(v, t),
      tclass_fn=lambda v, t: self._tclass_cast(ctx, v, t))


class _ExternalFunction:

  def __init__(self, fnname, fnmap, fnsig=None, dtype=None):
    marshals, kwmarshals = [], dict()
    if fnsig:
      for sarg in pyu.resplit(fnsig, ','):
        if m := re.match(r'\s*(\w+)\s*=\s*([^\s]+)', sarg):
          kwmarshals[m.group(1)] = _Marshal(m.group(2), fnname, m.group(1))
        else:
          marshals.append(_Marshal(sarg, fnname, len(marshals)))

    self._fnname = fnname
    self._fnmap = fnmap
    self._marshals = marshals
    self._kwmarshals = kwmarshals
    self._dtype = dtype
    # PyXHDL looks for __name__ to emit debug logging when running AST function calls.
    self.__name__ = fnname

  def __call__(self, *args, **kwargs):
    ctx = CodeGen.current()

    cargs = []
    for i, arg in enumerate(args):
      if i < len(self._marshals):
        arg = self._marshals[i](ctx, arg)

      cargs.append(arg)

    ckwargs = dict()
    for name, arg in kwargs.items():
      marshal = self._kwmarshals.get(name)
      if marshal is not None:
        arg = marshal(ctx, arg)

      ckwargs[name] = arg

    if isinstance(self._fnmap, dict):
      fmap = self._fnmap.get(ctx.emitter.kind)
      if fmap is None:
        fatal(f'Unable to resolve function {self._fnname}() for {ctx.emitter.kind} backend')
    else:
      fmap = self._fnmap

    dtype = self._dtype(cargs, ckwargs) if callable(self._dtype) else self._dtype

    if callable(fmap):
      call = fmap(ctx, cargs, ckwargs)

      return Value(dtype, call)
    else:
      if ckwargs:
        alog.warning(f'Keyword argument ignored for {self._fnname}() call: {ckwargs}')

      return ctx.emit_call(fmap, cargs, dtype)




import re

import py_misc_utils.core_utils as pycu
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .types import *
from .value_base import *


class VSpec:

  __slots__ = ('const', 'port', 'attributes')

  def __init__(self, const=False, port=None, attributes=None):
    self.const = const
    self.port = port
    self.attributes = attributes

  def __repr__(self):
    rfmt = pyu.repr_fmt(self, 'const,port,attributes')

    return f'{pyiu.cname(self)}({rfmt})'

  def __hash__(self):
    return hash((self.const, self.port, self.attributes))

  def __eq__(self, other):
    return (self.const == other.const and self.port == other.port and
            self.attributes == other.attributes)

  def for_new_variable(self):
    return pycu.new_with(self, port=None)


class Ref:

  __slots__ = ('name', 'mode', 'vspec', 'cname', 'vname')

  RW = 'RW'
  RO = 'RO'
  WO = 'WO'

  def __init__(self, name, mode=None, vspec=None, cname=None, vname=None):
    self.name = name
    self.mode = self._mode(mode, vspec)
    self.vspec = vspec
    self.cname = cname
    self.vname = vname

  @classmethod
  def _mode(cls, mode, vspec):
    if mode is None:
      mode = cls.RW if vspec is None or not vspec.const else cls.RO

    return mode

  def __repr__(self):
    rfmt = pyu.repr_fmt(self, 'name=,mode,vspec,cname,vname')

    return f'{pyiu.cname(self)}({rfmt})'

  def __hash__(self):
    return hash((self.name, self.mode, self.vspec, self.cname, self.vname))

  def __eq__(self, other):
    return (self.name == other.name and self.mode == other.mode and
            self.vspec == other.vspec and self.cname == other.cname and
            self.vname == other.vname)

  def new_name(self, name, vname=None):
    return pycu.new_with(self, name=name, vname=vname, cname=None)

  def new_mode(self, mode):
    return pycu.new_with(self, mode=mode)


class Init:

  __slots__ = ('value', 'name', 'vspec')

  def __init__(self, value=None, name=None, vspec=None):
    self.value = value
    self.name = name
    self.vspec = vspec

  def __hash__(self):
    return hash((self.value, self.name, self.vspec))

  def __eq__(self, other):
    return (self.value == other.value and self.name == other.name and
            self.vspec == other.vspec)

  def __repr__(self):
    rfmt = pyu.repr_fmt(self, 'value=,name,vspec')

    return f'{pyiu.cname(self)}({rfmt})'


class Value(ValueBase):

  __slots__ = ('dtype', '_value', 'isreg')

  def __init__(self, dtype, value, isreg=None):
    super().__init__()
    self.dtype = dtype
    self._value = value
    self.isreg = isreg

  @property
  def value(self):
    v = self._value
    return v.name if isinstance(v, Ref) else v

  @property
  def ref(self):
    v = self._value
    return v if isinstance(v, Ref) else None

  @property
  def vspec(self):
    v = self._value
    return v.vspec if isinstance(v, Ref) else None

  @property
  def init(self):
    v = self._value
    return v if isinstance(v, Init) else None

  @property
  def name(self):
    v = self._value
    return v.vname if isinstance(v, Ref) else None

  def is_none(self):
    return self._value is None

  def __repr__(self):
    rfmt = pyu.repr_fmt(self, '_value=,dtype,isreg')

    return f'{pyiu.cname(self)}({rfmt})'

  def __hash__(self):
    return hash((self.dtype, self._value, self.isreg))

  def __eq__(self, other):
    return (self.dtype == other.dtype and self._value == other._value and
            self.isreg == other.isreg)

  def new_value(self, value, shape=None, dtype=None, keepref=False):
    if dtype is None:
      vdtype = self.dtype.new_shape(*shape) if shape is not None else self.dtype
    else:
      vdtype = dtype if shape is None else dtype.new_shape(shape)

    if keepref and not isinstance(value, Ref):
      ref = self.ref
      if ref is not None:
        value = ref.new_name(value)

    return pycu.new_with(self, _value=value, dtype=vdtype)

  def new_isreg(self, isreg):
    return pycu.new_with(self, isreg=isreg)


class Wire(Value):

  def __init__(self, dtype, value):
    super().__init__(dtype, value, isreg=False)


class Register(Value):

  def __init__(self, dtype, value):
    super().__init__(dtype, value, isreg=True)


def valkind(isreg):
  return 'REG' if isreg is True else 'WIRE' if isreg is False else 'TEMP'


def _init_value(name, iargs):
  vspec = VSpec(**(iargs or dict()))

  return (Ref(name, vspec=vspec, cname=name, vname=name) if name is not None
          else Init(vspec=vspec))


def mkwire(dtype, name=None, **iargs) -> Value:
  return Wire(dtype, _init_value(name, iargs))


def mkreg(dtype, name=None, **iargs) -> Value:
  return Register(dtype, _init_value(name, iargs))


def mkvwire(dtype, value, name=None, **iargs) -> Value:
  vspec = VSpec(**(iargs or dict()))

  return Wire(dtype, Init(value=value, vspec=vspec, name=name))


def mkvreg(dtype, value, name=None, **iargs) -> Value:
  vspec = VSpec(**(iargs or dict()))

  return Register(dtype, Init(value=value, vspec=vspec, name=name))


def mknone(dtype) -> Value:
  return Register(dtype, None)


def is_ro_ref(v):
  vref = v.ref

  return vref is not None and vref.mode == Ref.RO


def has_hdl_vars(v):
  if isinstance(v, Value):
    return True
  elif isinstance(v, (list, tuple, set)):
    for e in v:
      if has_hdl_vars(e):
        return True
  elif pycu.isdict(v):
    for k, u in v.items():
      if has_hdl_vars(k) or has_hdl_vars(u):
        return True

  return False


def bitstring(value, remap=None) -> Value:
  # 0b0010 -> dtype=Bits(4), value="0010"
  m = re.match(r'0b([01XUZWHL]+)$', value)
  if m:
    bstr = ''.join(remap(x) for x in m.group(1)) if remap is not None else m.group(1)

    return Value(Bits(len(bstr)), bstr)


def bitfill(bitval, n):
  bitstr = str(bitval) if not isinstance(bitval, str) else bitval

  return '0b' + bitstr * n




import collections
import difflib
import inspect
import os
import sys
import unittest

import py_misc_utils.alog as alog
import py_misc_utils.core_utils as pycu

import pyxhdl as X
from pyxhdl import xlib as XL


_BACKEND_ARGS = collections.defaultdict(dict)


def add_backend_arg(backend, name, value):
  bdict = _BACKEND_ARGS[backend]
  pycu.dict_add(bdict, name, value)


def test_name(obj, name):
  func = getattr(obj, name)
  cls, name = func.__func__.__qualname__.split('.', 1)

  return f'{cls}__{name}'


def is_regen_mode():
  mode = os.environ.get('REGEN_TESTS')

  return mode is not None and int(mode) != 0


def data_folder():
  return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


_BACKEND_EXTS = {
  'vhdl': '.vhd',
  'verilog': '.sv',
}

def reference_path(name, backend, path=None):
  ext = _BACKEND_EXTS.get(backend, f'.{backend}')

  return os.path.join(path or data_folder(), f'{name}{ext}')


def load_reference(name, backend):
  path = reference_path(name, backend)
  if os.path.exists(path):
    with open(path, mode='r') as rfd:
      return rfd.read().split('\n')[: -1]


def store_reference(name, backend, code):
  path = reference_path(name, backend, path=os.environ.get('DEST_PATH'))
  alog.info(f'Storing reference code: name={name} banckend={backend} path={path}')
  with open(path, mode='w') as rfd:
    for cln in code:
      rfd.write(cln + '\n')


def generate_code(obj, inputs, backend):
  eargs = _BACKEND_ARGS[backend]
  emitter = X.Emitter.create(backend, **eargs)

  emitter.add_libpath(os.path.join(os.path.dirname(__file__), 'data', 'hdl_libs'))

  gglobals = X.create_globals(obj)
  codegen = X.CodeGen(emitter, gglobals)

  with codegen.context():
    if inspect.isclass(obj):
      codegen.generate_entity(obj, inputs)
    else:
      codegen.generate_process(obj, inputs)

    return codegen.flush()


def _run_test(test_obj, name, obj, inputs, backend):
  code = generate_code(obj, inputs, backend)

  ref_code = load_reference(name, backend)
  if ref_code is None:
    if is_regen_mode():
      store_reference(name, backend, code)
    else:
      test_obj.skipTest(f'Reference code not available for {backend} test: {name}')
  else:
    diff = difflib.unified_diff(ref_code, code,
                                fromfile=f'old_{name}',
                                tofile=f'new_{name}',
                                lineterm='')

    lines = list(diff)
    if lines:
      if is_regen_mode():
        store_reference(name, backend, code)
      else:
        txt_diff = '\n'.join(lines)
        test_obj.fail('\n' + txt_diff)


def run(test_obj, name, obj, inputs):
  for backend in X.Emitter.available():
    _run_test(test_obj, name, obj, inputs, backend)




import collections
import contextlib
import functools
import inspect
import os
import re
import sys

import py_misc_utils.alog as alog
import py_misc_utils.context_managers as pycm
import py_misc_utils.core_utils as pycu
import py_misc_utils.fs_utils as pyfsu
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.obj as obj
import py_misc_utils.ordered_set as pyos
import py_misc_utils.template_replace as pytr
import py_misc_utils.utils as pyu

from .common_defs import *
from .extern_logic import *
from .instantiator import *
from .types import *
from .vars import *
from .utils import *


OpSym = collections.namedtuple('OpSym', 'sym, isfn', defaults=[False])

FSpec = collections.namedtuple('FSpec', 'exp, mant')

_ProcessInfo = collections.namedtuple('ProcessInfo', 'name, kind, args, sens',
                                      defaults=[None, ROOT_PROCESS, None, None])

_Prec = collections.namedtuple('Prec', 'prec, prio, rtype', defaults=[0, None])

_ARITH_TYPE_PREC = {
  Bool: _Prec(-1, rtype=Uint(1)),
  Sint: _Prec(1, 2),
  Uint: _Prec(1, 1),
  Integer: _Prec(2),
  Float: _Prec(3),
  Real: _Prec(4),
}

_BIT_TYPE_PREC = {
  Bool: _Prec(-1, rtype=Uint(1)),
  Sint: _Prec(1, 1),
  Uint: _Prec(1, 2),
  Bits: _Prec(1, 3),
}

_CONCAT_TYPE_PREC = {
  Bool: _Prec(-1, rtype=Bits(1)),
  Sint: _Prec(1),
  Uint: _Prec(2),
  Bits: _Prec(3),
}

_COMPARE_TYPE_PREC = {
  Bool: _Prec(-1),
  Bits: _Prec(1),
  Sint: _Prec(2, 2),
  Uint: _Prec(2, 1),
  Integer: _Prec(3),
  Float: _Prec(4),
  Real: _Prec(5),
}

_IFEXP_TYPE_PREC = {
  Bool: _Prec(1),
  Sint: _Prec(2, 2),
  Uint: _Prec(2, 1),
  Integer: _Prec(3),
  Float: _Prec(4),
  Real: _Prec(5),
  Bits: _Prec(6),
}

_CTYPES_ALLOWED = {
  float: 1,
  int: 2,
  str: 3,
}

_FLOAT_SPECS = {
  16: FSpec(5, 10),
  32: FSpec(8, 23),
  64: FSpec(11, 52),
  80: FSpec(17, 63),
  128: FSpec(15, 112),
}

class _Placement:

  def __init__(self, indent=None):
    self.code = []
    self.indent = indent

  def append(self, code):
    self.code.append(code)

  def __len__(self):
    count = 0
    for c in self.code:
      if isinstance(c, _Placement):
        count += len(c)
      else:
        count += 1

    return count


class Emitter:

  _BACKEND_REGISTRY = dict()
  _MODULE_REGISTRY = collections.defaultdict(dict)

  def __init__(self, cfg_file=None, **kwargs):
    self._cfg_file = cfg_file
    self._cfg = pyu.load_config(cfg_file, extra=kwargs) if cfg_file else dict()
    self._indent_spaces = self._cfg.get('indent_spaces', 2)
    self._indent = 0
    self._placements = []
    self._code = []
    self._extra_libs = pyos.OrderedSet()
    self._lib_paths = []
    self._ent_versions = collections.defaultdict(int)
    self._user_modules = collections.defaultdict(dict)
    self._extern_modules = dict()
    self._contexts = []
    self._module_reset()

  @classmethod
  def register(cls, name, eclass):
    cls._BACKEND_REGISTRY[name] = eclass

  @classmethod
  def available(cls):
    return sorted(cls._BACKEND_REGISTRY.keys())

  @classmethod
  def create(cls, name, **kwargs):
    eclass = cls._BACKEND_REGISTRY.get(name)
    if eclass is None:
      fatal(f'Unknown emitter: {name}')

    return eclass(**kwargs)

  @classmethod
  def _register_module(cls, mid, code, dest, replace):
    for kind, kcode in code.items():
      mdict = dest[kind]
      if mid in mdict:
        if replace is False:
          fatal(f'Module "{mid}" already registered')
        else:
          alog.info(f'Module "{mid}" already registered, overriding!')

      mdict[mid] = kcode.split('\n')

  @classmethod
  def glob_register_module(cls, mid, code, replace=None):
    cls._register_module(mid, code, cls._MODULE_REGISTRY, replace=replace)

  def add_libpath(self, path):
    self._lib_paths.append(pyfsu.normpath(path))

  def add_extra_library(self, name):
    self._extra_libs.add(name)

  def register_module(self, mid, code, replace=None):
    self._register_module(mid, code, self._user_modules, replace=replace)

  @staticmethod
  def xmod_resolve(modname, fnname, argref, **kwargs):
    def resolver(ctx, cargs, rkwargs):
      xkwargs = kwargs.copy()
      xkwargs.update(rkwargs)
      return ctx.emitter._xmod_resolve(modname, fnname, cargs, argref, xkwargs)

    return resolver

  def _argref_dtype(self, argref, cargs, kwargs):
    if isinstance(argref, Type):
      return argref

    arg_value = cargs[argref] if isinstance(argref, int) else kwargs[argref]

    return arg_value.dtype

  def _xmod_resolve(self, modname, fnname, cargs, argref, kwargs):
    xmod = self._get_extern_module(modname)

    dtype = self._argref_dtype(argref, cargs, kwargs)

    return self._call_external_module(xmod, fnname, dtype, *cargs, **kwargs)

  def load_extern_module(self, path):
    if os.path.isabs(path):
      mod_path = path
    else:
      mod_path = os.path.join(os.path.dirname(__file__), path)

    xmod = ExternModule.load(mod_path)

    self._extern_modules[xmod.name] = xmod

    return xmod

  def _get_extern_module(self, name):
    xmod = self._extern_modules.get(name)
    if xmod is None:
      fatal(f'Extern module {name} is not available')

    return xmod

  def _module_arg(self, modname, name):
    return f'{modname}.{name}'

  def _call_external_module(self, xmod, fnname, dtype, *args, **kwargs):
    xlogic = xmod.get_logic(fnname)

    if xlogic.nargs != len(args):
      fatal(f'Number of arguments mismatch for {fnname}: {xlogic.nargs} vs. {len(args)}')

    self.add_extra_library(xlogic.filename)

    mod_params, mod_args = dict(), dict()
    if isinstance(xlogic.params, (list, tuple)):
      for pname in xlogic.params:
        pvalue = kwargs.get(pname)
        if pvalue is None:
          fatal(f'Parameter {pname} missing when using function {fnname} of ' \
                f'external module {xmod.name}')

        rpname = xlogic.name_remap.get(pname, pname)
        mod_params[rpname] = pvalue
    else:
      for pname, value in xlogic.params.items():
        pvalue = kwargs.get(pname, value)
        rpname = xlogic.name_remap.get(pname, pname)
        mod_params[rpname] = pvalue

    result = None
    for aname, aref in xlogic.args.items():
      if m := re.match(r'\$!([^=]+)=(.*)$', aref):
        raname = xlogic.name_remap.get(aname, aname)
        dtype = dtype_from_string(m.group(1))
        avalue = pycu.infer_value(m.group(2))
        mod_args[raname] = self.cast(avalue, dtype)
      elif m := re.match(r'\$#(\d+)=(.*)$', aref):
        raname = xlogic.name_remap.get(aname, aname)
        argno = int(m.group(1))
        avalue = pycu.infer_value(m.group(2))
        mod_args[raname] = self.cast(avalue, args[argno].dtype)
      elif m := re.match(r'\$(\w+)=(.*)$', aref):
        raname = xlogic.name_remap.get(aname, aname)
        avalue = pycu.infer_value(m.group(2))
        mod_args[raname] = self.cast(avalue, kwargs[m.group(1)].dtype)
      else:
        raname = xlogic.name_remap.get(aref, aref)
        if m := re.match(r'\$(\d+)$', aname):
          argno = int(m.group(1))
          mod_args[raname] = args[argno]
        elif aname == '$RESULT':
          # Lazy import to avoid cycles.
          from . import xlib

          resname = xlib.generate_name(fnname)
          xlib.assign(resname, mkreg(dtype, name=resname))
          mod_args[raname] = result = xlib.load(resname)
        else:
          avalue = kwargs.get(aname)
          if avalue is None:
            fatal(f'Argument {aname} missing when using function {fnname} of ' \
                  f'external module {xmod.name}')

          mod_args[raname] = avalue

    mod_args[PARAM_KEY] = mod_params

    if xlogic.funcname:
      iid = self._itor.getid(xlogic.modname, mod_args)

      mcall = self._module_arg(iid, xlogic.funcname)

      return f'{mcall}(' + ', '.join(self.svalue(arg) for arg in args) + ')'
    else:
      iid = self._itor.getid(xlogic.modname, mod_args, force_new=True)

      return result.ref

  def float_spec(self, dtype):
    fsenv = os.getenv(f'F{dtype.nbits}_SPEC')
    if fsenv is not None:
      return FSpec(int(x) for x in pyu.resplit(fsenv, ','))

    fspecs = self._cfg.get('float_specs', _FLOAT_SPECS)
    fspec = fspecs.get(dtype.nbits)
    if fspec is None:
      fatal(f'Unknown floating point spec: {dtype.nbits} bits',
            exc=TypeError)

    return fspec if isinstance(fspec, FSpec) else FSpec(**fspec)

  def get_contexts(self, kind):
    all_contexts = []
    for ctx in self._contexts:
      kctx = ctx.get(kind)
      if kctx is not None:
        all_contexts.append(kctx)

    return all_contexts

  def get_context(self, kind, defval=None):
    ctxs = self.get_contexts(kind)

    return ctxs[-1] if ctxs else defval

  def create_placement(self, extra_indent=0):
    return _Placement(indent=self._indent + extra_indent)

  def emit_placement(self, placement=None, extra_indent=0):
    place = self.create_placement(extra_indent=extra_indent)
    self._emit(place, placement=placement or self.curr_placement())

    return place

  def svalue(self, value):
    xvalue = value.value if isinstance(value, Value) else value

    mvalue = self._scalar_remap(xvalue)
    if mvalue is not None:
      xvalue = mvalue
    elif not isinstance(xvalue, str):
      xvalue = str(xvalue)

    return xvalue

  def _emit(self, obj, placement=None):
    if placement is None:
      self._code.append(obj)
    else:
      placement.append(obj)

  def _emit_line(self, line):
    spaces = ' ' * (self._indent_spaces * self._indent)
    if not self._placements:
      self._emit(spaces + line)
    else:
      self._emit(spaces + line, placement=self._placements[-1])

  def _emit_lines(self, lines, sep=''):
    if not hasattr(lines, '__len__'):
      lines = tuple(lines)

    count = len(lines)
    for i, ln in enumerate(lines):
      if i + 1 < count:
        self._emit_line(ln + sep)
      else:
        self._emit_line(ln)

  def _expand_helper(self, code, lines):
    for ent in code:
      if isinstance(ent, _Placement):
        self._expand_helper(ent.code, lines)
      elif isinstance(ent, (list, tuple)):
        self._expand_helper(ent, lines)
      elif callable(ent):
        lines.extend(ent(size=len(code)))
      else:
        lines.append(ent)

    return lines

  def _expand(self):
    return self._expand_helper(self._code, [])

  def _process_init(self, name, kind, args, sens):
    self._proc = _ProcessInfo(name=name, kind=kind, args=args, sens=sens)

  def _process_reset(self):
    self._proc = _ProcessInfo()

  def _module_reset(self):
    self._mod_comment = None
    self._itor = Instanciator(param_key=PARAM_KEY)
    self._revgen = pycu.RevGen(fmt='{name}{ver}')
    self._process_reset()

  def _int2bits(self, intval, n):
    value = intval & ((1 << n) - 1)

    return f'{value:0{n}b}'

  def _paren_join(self, joiner, args):
    return joiner.join(paren(x) for x in args) if len(args) > 1 else args[0]

  def _default_float_type(self):
    return dtype_from_string(os.getenv('FLOAT_TYPE', 'f32'))

  def _match_intstring(self, value):
    dtype = ivalue = None
    # u16`1234 -> dtype=Uint(16), ivalue=1234
    m = re.match(r'([us])(\d+)`(.*)$', value)
    if m:
      nbits = int(m.group(2))
      ivalue = pycu.infer_value(m.group(3))
      if not isinstance(ivalue, int):
        fatal(f'Invalid literal value: {value}')

      if m.group(1) == 'u':
        dtype = Uint(nbits)
      else:
        dtype = Sint(nbits)

    return dtype, ivalue

  def _convert_int(self, value, tclass=None):
    if tclass is None:
      tclass = Uint if value >= 0 else Sint

    return Value(tclass(value.bit_length()), str(value))

  def _best_type(self, cur, new, type_prec):
    nprec = type_prec.get(type(new))
    if nprec is None:
      fatal(f'Unsupported type {pyiu.cname(new)} ... should be ' \
            f'{tuple(pyiu.cname(x) for x in type_prec.keys())}')

    if cur is not None:
      cprec = type_prec[type(cur)]
      dprec = nprec.prec - cprec.prec
      if dprec > 0:
        return new
      elif dprec < 0:
        return cur
      else:
        nnbits = sys.maxsize if new.nbits is None else new.nbits
        cnbits = sys.maxsize if cur.nbits is None else cur.nbits
        if nnbits > cnbits:
          return new
        elif nnbits < cnbits:
          return cur
        else:
          return new if nprec.prio > cprec.prio else cur

    return new

  def _result_type(self, dtype, type_prec):
    if dtype is not None:
      prec = type_prec.get(type(dtype))
      if prec is not None and prec.rtype is not None:
        dtype = prec.rtype

    return dtype

  def _marshal_arith_op(self, args, ctype_allowed=None, ctype_cast=None):
    return self._gen_marshal(args, _ARITH_TYPE_PREC, ctype_allowed or _CTYPES_ALLOWED,
                             ctype_cast=ctype_cast)

  def _marshal_concat_op(self, args):
    dtype, margs, nbits = None, [], 0
    for arg in args:
      if not isinstance(arg, Value):
        arg = self._try_convert_literal(arg)
        if not isinstance(arg, Value):
          pyu.assert_instance('Integer type required', arg, int)
          arg = self._convert_int(arg)

      margs.append(arg)
      nbits += arg.dtype.nbits
      dtype = self._best_type(dtype, arg.dtype, _CONCAT_TYPE_PREC)

    dtype = self._result_type(dtype, _CONCAT_TYPE_PREC)

    tclass = type(dtype)
    for i, arg in enumerate(margs):
      if type(arg.dtype) != tclass:
        atype = tclass(arg.dtype.nbits)
        margs[i] = Value(atype, self._cast(arg, atype))

    return tclass(nbits), margs

  def _marshal_bit_op(self, args):
    dtype, margs = None, []
    for arg in args:
      if not isinstance(arg, Value):
        arg = self._try_convert_literal(arg)
      margs.append(arg)
      if isinstance(arg, Value):
        dtype = self._best_type(dtype, arg.dtype, _BIT_TYPE_PREC)
      else:
        pyu.assert_instance('Integer type required', arg, int)

    dtype = self._result_type(dtype, _BIT_TYPE_PREC)
    if dtype is None:
      nbits = max(arg.bit_length() for arg in args)
      dtype = Uint(nbits)

    return [Value(dtype, self._cast(arg, dtype)) for arg in margs]

  def _marshal_shift_op(self, left, right):
    if isinstance(right, Value):
      if not isinstance(right.dtype, Integer):
        rdtype = Integer()
        right = Value(rdtype, self._cast(right, rdtype))
    elif type(right) != int:
      fatal(f'Shift amount should be an integer: {right}')
    if not isinstance(left, Value):
      left = self._try_convert_literal(arg)
      if not isinstance(left, Value):
        pyu.assert_instance('Shift operand should be an integer', left, int)
        dtype = Uint(left.bit_length())
        left = Value(dtype, self._cast(left, dtype))
      else:
        pyu.assert_instance('Unexpected type', left.dtype, (Bits, Sint, Uint, Integer))

    return left, right

  def _marshal_compare_op(self, args, ctype_allowed=None, ctype_cast=None):
    return self._gen_marshal(args, _COMPARE_TYPE_PREC, ctype_allowed or _CTYPES_ALLOWED,
                             ctype_cast=ctype_cast)

  def _marshal_ifexp_op(self, args, ctype_allowed=None, ctype_cast=None):
    return self._gen_marshal(args, _IFEXP_TYPE_PREC, ctype_allowed or _CTYPES_ALLOWED,
                             ctype_cast=ctype_cast)

  def _gen_marshal(self, args, type_prec, ctype_allowed, ctype_cast=None):
    cargs, dtype, ctype = [], None, None
    for arg in args:
      if not isinstance(arg, Value):
        arg = self._try_convert_literal(arg)
      cargs.append(arg)
      if isinstance(arg, Value):
        dtype = self._best_type(dtype, arg.dtype, type_prec)
      else:
        ytype = type(arg)
        yind = ctype_allowed.get(ytype, -1)
        if yind < 0:
          fatal(f'Type {ytype} not allowed, should be one of ' \
                f'{tuple(pyiu.cname(t) for t in ctype_allowed.keys())}')
        if ctype is None or ctype_allowed[ctype] > yind:
          ctype = ytype

    dtype = self._result_type(dtype, type_prec)
    if dtype is None:
      if ctype == float:
        dtype = Real()
      elif ctype == int:
        dtype = Integer()
      else:
        fatal(f'Unknown type: {ctype}')

    for i, arg in enumerate(cargs):
      if isinstance(arg, Value):
        if dtype != arg.dtype:
          cargs[i] = Value(dtype, self._cast(arg, dtype))
      else:
        cargs[i] = (Value(dtype, self._cast(arg, dtype)) if ctype_cast is None else
                    ctype_cast(arg, dtype))

    return cargs

  def _cfg_lookup(self, k, defval=None):
    v = pyu.dict_rget(self._cfg, f'env/{k}')
    if v is None:
      v = os.getenv(f'PYXHDL_{k}')
      if v is None:
        if defval is None:
          fatal(f'Missing configuration: {k}')
        else:
          v = defval

    return v

  def _load_code(self, path):
    with open(path, mode='r') as fd:
      code = fd.read()

    return pytr.template_replace(code, lookup_fn=self._cfg_lookup, delim='@')

  def _collect_libpaths(self):
    libdir = os.path.join(os.path.dirname(__file__), 'hdl_libs', self.kind)

    lib_paths = [libdir] + [os.path.join(path, self.kind) for path in self._lib_paths]

    lib_paths.extend(pyfsu.normpath(x)
                     for x in self._cfg.get('lib_paths', dict()).get(self.kind, ()))

    if env_paths := os.getenv(f'PYXHDL_{self.kind.upper()}_LIBPATH'):
      lib_paths.extend(pyfsu.normpath(x) for x in pyu.resplit(env_paths, ';'))

    alog.debug(f'Using library folders {lib_paths}')

    return lib_paths

  def _load_libs(self):
    lib_paths = self._collect_libpaths()

    libcode, xlibs = [], []

    # Loading internal libraries by manifest (always load ones).
    if lpath := pyfsu.find_path('LIBS', lib_paths):
      with open(lpath, mode='r') as mfd:
        for libname in [l.strip() for l in mfd.read().split('\n')]:
          if libname and not libname.startswith('#'):
            xlibs.append(libname)

    xlibs.extend(self._extra_libs)

    if env_libs := os.getenv(f'PYXHDL_{self.kind.upper()}_LIBS'):
      xlibs.extend(pyu.resplit(env_libs, ';'))

    xlibs.extend(self._cfg.get('libs', dict()).get(self.kind, ()))

    for libname in pycu.enum_unique(xlibs):
      _, ext = os.path.splitext(libname)

      libfname = (libname + self.file_ext) if not ext else libname

      if lpath := pyfsu.find_path(libfname, lib_paths):
        alog.debug(f'Loading {self.kind} library file {lpath}')
        libcode.extend(self._load_code(lpath).split('\n'))
      else:
        fatal(f'Library "{libname}" ("{libfname}") not found in {lib_paths}')

    # Add user defined modules within the PyXHDL code.
    for cid, umod in self._MODULE_REGISTRY[self.kind].items():
      libcode.extend(umod)
    for cid, umod in self._user_modules[self.kind].items():
      libcode.extend(umod)

    return libcode

  def _get_entity_inst(self, name):
    self._ent_versions[name] += 1

    return f'{name}_{self._ent_versions[name]}'

  def curr_placement(self):
    return self._placements[-1] if self._placements else None

  @contextlib.contextmanager
  def context(self, ctx):
    self._contexts.append(ctx)
    try:
      yield self
    finally:
      self._contexts.pop()

  @contextlib.contextmanager
  def placement(self, place):
    self._placements.append(place)
    indent, self._indent = self._indent, place.indent
    try:
      yield self
    finally:
      self._indent = indent
      self._placements.pop()

  @contextlib.contextmanager
  def indent(self):
    self._indent += 1
    try:
      yield self
    finally:
      self._indent -= 1

  def emit_code(self, code):
    for ln in code.split('\n'):
      self._emit_line(ln)

  def build_args_string(self, fn, delim, args):
    sargs = [fn(self.svalue(a)) for a in args]

    return delim.join(sargs)

  def emit_call(self, fname, args, dtype):
    cargs = self.build_args_string(lambda a: a, ', ', args)

    call = f'{fname}({cargs})'

    alog.debug(lambda: f'{call} -> {dtype}')

    # None return type means no return value call, which is emitted directly.
    # Otherwise it is a function, which returns a Value, which materializes only
    # when/if used (ie, within an expression or as RHS of an assignment).
    # This means that if an external function has side effects (ie, output arguments)
    # and a return value, such return value must be used (assigned to something or
    # used within an expression) in order to the side effects to manifest.
    if dtype is None:
      self.emit_code(f'{call}{self.eol}')
    else:
      return Value(dtype, call)

  def tclass_cast(self, tclass, value):
    # The input `value` here is a Python string literal, not a Value.
    if tclass in (Sint, Uint):
      value = int(value)
      dtype = tclass(value.bit_length())
    elif tclass == Bits:
      if isinstance(value, int):
        nbits = value.bit_length()
      else:
        nbits = len(value) - (2 if value.startswith('0b') else 0)
      dtype = Bits(nbits)
    elif tclass == Float:
      dtype = self._default_float_type()
    elif tclass in (Bool, Integer, Real):
      dtype = tclass()
    else:
      fatal(f'Unsupported class "{pyiu.cname(tclass)}" while converting "{value}"')

    return self.cast(value, dtype)

  def time_unit(self):
    return self._cfg_lookup('TIME_UNIT', defval='ns')

  def _normalize_time(self, ts):
    tu = self.time_unit()

    return scaled_time(ts, tu), tu

  def _enum_attributes(self, attributes):
    for aname in ('$common', self.kind):
      attrs = attributes.get(aname)
      if attrs is not None:
        for name, value in attrs.items():
          yield name, value

  def _codegen_ctx(self):
    from . import pyxhdl as X

    return X.CodeGen.current()




# Cannot depend of local modules!
import collections
import contextlib
import copy
import inspect
import os
import re
import sys
import textwrap

import py_misc_utils.core_utils as pycu
import py_misc_utils.utils as pyu


_FuncInfo = collections.namedtuple('FuncInfo', 'filename, lineno, source')

class _None:

  def __repr__(self):
    return 'NONE'


NONE = _None()


def get_obj_globals(obj, defval=None):
  mod = getattr(obj, '__module__', None)
  if mod is not None:
    imod = sys.modules.get(mod)
    if imod is not None:
      return imod.__dict__

  return getattr(obj, '__globals__', defval)


def create_globals(obj, source_globals=None):
  obj_globs = get_obj_globals(obj, defval=dict())
  if source_globals:
    gglobals = copy.copy(source_globals)
    for k, v in pycu.enum_values(obj_globs):
      if not k.startswith('__'):
        gglobals[k] = v
  else:
    gglobals = {k: v for k, v in pycu.enum_values(obj_globs)}

  return gglobals


def vload(name, globs, locs):
  v = pyu.getvar(locs, name, NONE)
  if v is NONE:
    v = pyu.getvar(globs, name, NONE)
    if v is NONE:
      bins = globs.get('__builtins__')
      if bins is not None:
        v = pyu.getvar(bins, name, NONE)

  return v


# Do not add "'" as this needs not to be parenthesized in SystemVerilog.
_BOUNDS = {
  '{': '}',
  '(': ')',
  '[': ']',
  '"': '"',
}
_NEEDSPAR_RX = re.compile(r'[ +\-*/%!~&|^]')

def paren(sx, kind='()'):
  level, levch, skip = 0, [None], False
  for c in sx:
    if skip:
      skip = False
    elif c == '\\':
      skip = True
    elif c == levch[-1]:
      levch.pop()
      level -= 1
    elif c in _BOUNDS:
      levch.append(_BOUNDS[c])
      level += 1
    elif level == 0 and _NEEDSPAR_RX.match(c):
      return f'{kind[0]}{sx}{kind[1]}'

  return sx


_FN_INFO = '__fninfo__'

def get_function_info(func):
  fninfo = getattr(func, _FN_INFO, None)
  if fninfo is None:
    filename = inspect.getsourcefile(func)
    slines, lineno = inspect.getsourcelines(func)
    source = textwrap.dedent(''.join(slines))

    fninfo = _FuncInfo(filename=filename, lineno=lineno, source=source)

  return fninfo


def set_function_info(func, filename, lineno, source):
  setattr(func, _FN_INFO, _FuncInfo(filename=filename, lineno=lineno, source=source))


def needs_self(func):
  sig = inspect.signature(func)
  if sig.parameters:
    name, param = next(iter(sig.parameters.items()))

    return name == 'self'

  return False


def flat2shape(parts, shape, opar, cpar):
  sparts = parts
  for dim in reversed(shape):
    elist = [sparts[x: x + dim] for x in range(0, len(sparts), dim)]
    sparts = [opar + ', '.join(l) + cpar for l in elist]

  return sparts[0]


class _EntityRecord:

  def __init__(self, eclass, pargs, kwargs):
    self.eclass = eclass
    self.pargs = pargs
    self.kwargs = kwargs
    self.hash = pycu.genhash((eclass, pargs, kwargs))

  def __hash__(self):
    return self.hash

  def __eq__(self, other):
    return (self.eclass == other.eclass and self.kwargs == other.kwargs and
            self.pargs == other.pargs)


class EntityVersions:

  def __init__(self):
    self._versions = collections.defaultdict(dict)

  def getname(self, eclass, pargs, kwargs):
    erec = _EntityRecord(eclass, pargs, kwargs)

    ent_name = eclass.__name__
    eldict = self._versions[ent_name]
    ver = eldict.get(erec)
    if ver is None:
      ver = len(eldict)
      eldict[erec] = ver

    return ent_name if ver == 0 else f'{ent_name}_{ver}', erec


def subscript_setter(arr, idx):

  def setfn(value):
    try:
      arr[idx] = value
    except Exception as ex:
      raise pycu.rewrited_exception(ex, f'; arr = {arr}, idx = {idx}') from None

  return setfn


@contextlib.contextmanager
def temp_attributes(obj, **attrs):
  saved = dict()
  for k, v in attrs.items():
    saved[k] = getattr(obj, k, NONE)
    setattr(obj, k, v)

  try:
    yield obj
  finally:
    for k, v in saved.items():
      if v is NONE:
        delattr(obj, k)
      else:
        setattr(obj, k, v)


def subname(*path):
  return '_'.join(path)


_TIME_SCALES = {'s': 1.0, 'ms': 1e3, 'us': 1e6, 'ns': 1e9, 'ps': 1e12}
_TIME_UNITS = '|'.join(_TIME_SCALES.keys())
_TIME_REX = re.compile(rf'([+\-]?\d+(\.(\d*))?([eE][+\-]?\d+)?)\s*({_TIME_UNITS})?$')

def scaled_time(ts, tu):
  scale = _TIME_SCALES.get(tu.lower())
  if scale is None:
    fatal(f'Invalid time unit: {tu}')

  return ts * scale


def parse_time(ts):
  m = _TIME_REX.match(ts)
  if m is None:
    fatal(f'Invalid time format: {ts}')

  nts = float(m.group(1))

  return nts / _TIME_SCALES[m.group(5)] if m.group(5) else nts


def fatal(msg, exc=RuntimeError):
  pyu.fatal(msg, exc=exc)


def validate_slice(start, stop, size, step):
  nstart, nstop = pycu.norm_slice(start, stop, size)

  # Slices must be at least of size one (empty slices are not allowed).
  if step > 0:
    if nstart < 0 or nstart >= nstop or nstop > size:
      fatal(f'Invalid slice: start={nstart} stop={nstop} step={step}', exc=ValueError)
  elif step < 0:
    if nstart >= size or nstop >= nstart or nstop < -1:
      fatal(f'Invalid slice: start={nstart} stop={nstop} step={step}', exc=ValueError)
  else:
    fatal(f'Invalid slice: start={nstart} stop={nstop} step={step}', exc=ValueError)

  return nstart, nstop




import ast
import copy

import py_misc_utils.ast_utils as asu

from .utils import *


def ast_hdl_transform(node):
  asu.ifize(node)

  return node


def elements(c):
  if isinstance(c, (list, tuple)):
    return c
  if isinstance(c, (ast.List, ast.Tuple)):
    return c.elts


def as_loading(node, inplace=False):

  class Transformer(ast.NodeTransformer):

    def visit_Store(self, node):
      return ast.Load()


  cnode = node if inplace else copy.deepcopy(node)

  return Transformer().visit(cnode)




import ast

import py_misc_utils.lazy_import as pyli
import py_misc_utils.utils as pyu

X = pyli.lazy_import('pyxhdl')


# This is the Value's base class and implements a Python data model for the Value class.
# NOTE: This interface gets only used when performing Value instances operations outside
# of the HDL marked scope (the @X.hdl or @X.hdl_process(...) decorators).
# Note also that although it is possible to perform certain operations of Value while outside
# of the HDL scopes, a few things do not work.
# It is not possible to use Python IF statements depending on Value instances, as well as it
# is not possible to compose logical operations on them (like, "a > b and e < f" - only
# possible using the ValueBase And(), Or() and Not() functions, like "(a > b).And(e < f)").
# To handle those cases, PyXHDL has to interpret the AST tree and emit the matching
# HDL code, which happens while within an HDL scope.
# IOW, HDL variables (Value instances) processing should happen within HDL scopes, and
# the datamodel interface is only provided as fallback.
class ValueBase:

  @staticmethod
  def load(name):
    ctx = X.CodeGen.current()

    return ctx.load_var(name)

  def store(self, name):
    ctx = X.CodeGen.current()

    ctx.assign_value(name, self)

  # Conversions
  def cast(self, dtype):
    ctx = X.CodeGen.current()

    return ctx.emitter.cast(self, dtype)

  # Arithmetic
  def __add__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Add(), self, other)

  def __sub__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Sub(), self, other)

  def __mul__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Mult(), self, other)

  def __truediv__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Div(), self, other)

  def __mod__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Mod(), self, other)

  def __lshift__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.LShift(), self, other)

  def __rshift__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.RShift(), self, other)

  def __and__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.BitAnd(), self, other)

  def __or__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.BitOr(), self, other)

  def __xor__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.BitXor(), self, other)

  def __matmul__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.MatMult(), self, other)

  # Unaries
  def __neg__(self):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_UnaryOp(ast.USub(), self)

  def __pos__(self):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_UnaryOp(ast.UAdd(), self)

  def __invert__(self):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_UnaryOp(ast.Invert(), self)

  # Compare
  def __eq__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_Compare(self, [ast.Eq()], [other])

  def __ne__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_Compare(self, [ast.NotEq()], [other])

  def __lt__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_Compare(self, [ast.Lt()], [other])

  def __le__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_Compare(self, [ast.LtE()], [other])

  def __gt__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_Compare(self, [ast.Gt()], [other])

  def __ge__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_Compare(self, [ast.GtE()], [other])

  # The *R* versions
  def __radd__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Add(), other, self)

  def __rsub__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Sub(), other, self)

  def __rmul__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Mult(), other, self)

  def __rtruediv__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Div(), other, self)

  def __rmod__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.Mod(), other, self)

  def __rlshift__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.LShift(), other, self)

  def __rrshift__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.RShift(), other, self)

  def __rand__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.BitAnd(), other, self)

  def __ror__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.BitOr(), other, self)

  def __rxor__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.BitXor(), other, self)

  def __rmatmul__(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BinOp(ast.MatMult(), other, self)

  # Arithmetic Self Assign
  def _ires(self, res):
    return pyu.copy_inplace(self, res)

  def __iadd__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.Add(), self, other))

  def __isub__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.Sub(), self, other))

  def __imul__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.Mult(), self, other))

  def __itruediv__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.Div(), self, other))

  def __imod__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.Mod(), self, other))

  def __ilshift__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.LShift(), self, other))

  def __irshift__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.RShift(), self, other))

  def __iand__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.BitAnd(), self, other))

  def __ior__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.BitOr(), self, other))

  def __ixor__(self, other):
    ctx = X.CodeGen.current()

    return self._ires(ctx.emitter.eval_BinOp(ast.BitXor(), self, other))

  # Item operations
  def __getitem__(self, idx):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_Subscript(self, idx)

  # Logical **functions** (note that it is not possible to override the Python
  # logical AND, OR and NOT at the datamodel level).
  def And(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BoolOp(ast.And(), (self, other))

  def Or(self, other):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_BoolOp(ast.Or(), (self, other))

  def Not(self):
    ctx = X.CodeGen.current()

    return ctx.emitter.eval_UnaryOp(ast.Not(), self)

  # Allow for Value functions to map directly to emitter eval_*() ones.
  def __getattr__(self, attr):
    ctx = X.CodeGen.current()

    emitter_fn = getattr(ctx.emitter, f'eval_{attr}', None)
    if emitter_fn is None:
      from . import utils

      utils.fatal(f'Not implemented: {attr}', exc=AttributeError)

    def fncall(*args, **kwargs):
      return emitter_fn(self, *args, **kwargs)

    return fncall




import ast
import collections
import contextlib
import copy
import enum
import functools
import inspect
import re
import textwrap

import py_misc_utils.alog as alog
import py_misc_utils.ast_utils as asu
import py_misc_utils.core_utils as pycu
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.thread_context as pytc
import py_misc_utils.utils as pyu

from .ast_utils import *
from .common_defs import *
from .decorators import *
from .entity import *
from .interface import *
from .types import *
from .vars import *
from .utils import *
from .wrap import *


_LoopContext = collections.namedtuple('LoopContext', 'mode')
_ForLoop = collections.namedtuple('ForLoop', 'data, ivar, start, end, step',
                                  defaults=(None, None, None, None))
_Return = collections.namedtuple('Return', 'value, placement')
_MatchCase = collections.namedtuple('MatchCase', 'pattern, scope')

_CGENCTX = 'pyxhdl.CodeGen'
_CODEFMT_RX = r'(?<!\{)\{([^{][^}]*(\}\}[^}]+)*)\}'


class Variable:

  __slots__ = ('dtype', 'isreg', 'init', 'vspec')

  def __init__(self, dtype, isreg, init=None, vspec=None):
    self.dtype = dtype
    self.isreg = isreg
    self.init = init
    self.vspec = vspec

  def is_const(self):
    return self.vspec is not None and self.vspec.const

  def __repr__(self):
    return f'{pyiu.cname(self)}({self.dtype}, {self.isreg}, {self.init}, {self.vspec})'


class _LoopModes(enum.IntEnum):
  UNROLLED = enum.auto()
  HDL = enum.auto()


class _Exception(Exception):
  pass


class _ReturnException(_Exception):

  def __init__(self, value):
    super().__init__()
    self.value = value


class _BreakException(_Exception):
  pass


class _ContinueException(_Exception):
  pass


class _SourceLocation:

  def __init__(self, filename, base_lineno):
    self.filename = filename
    self.base_lineno = base_lineno
    self.lineno = base_lineno

  def set_lineno(self, lineno):
    self.lineno = self.base_lineno + lineno - 1


class _Storer:

  def __init__(self, fn):
    self.fn = fn

  def store(self, value):
    self.fn(value)


class _HdlChecker(ast.NodeVisitor):

  HDL_TYPES = (Value, Interface, InterfaceView)

  def __init__(self, globs, locs, hdl_required=('Yield', 'YieldFrom')):
    super().__init__()
    self._globs = globs
    self._locs = locs
    self._scope = 0
    self._stack = []
    self.count = 0
    for name in pyu.expand_strings(hdl_required):
      setattr(self, f'visit_{name}', self._checker)

  def _checker(self, node):
    self.count += 1

  @contextlib.contextmanager
  def _scope_in(self):
    self._scope += 1
    try:
      yield self
    finally:
      self._scope -= 1
      if self._scope == 0:
        self._stack = []

  def _pop_value(self):
    return self._stack.pop() if self._stack else NONE

  def _push_value(self, value):
    if self._scope > 0:
      self._stack.append(value)
    if isinstance(value, self.HDL_TYPES):
      self.count += 1

  def _needs_hdl_call(self, func):
    if is_hdl_function(func):
      return True
    if inspect.isfunction(func):
      sig = inspect.signature(func)
      if sig.return_annotation in self.HDL_TYPES:
        return True
    if not inspect.isclass(func):
      return False

    return (pyiu.is_subclass(func, Entity) or func in self.HDL_TYPES or
            is_hdl_function(getattr(func, '__init__', None)))

  def visit_Attribute(self, node):
    with self._scope_in():
      self.visit(node.value)
      if (avalue := getattr(self._pop_value(), node.attr, NONE)) is not NONE:
        self._push_value(avalue)

  def visit_Name(self, node):
    value = vload(node.id, self._globs, self._locs)
    self._push_value(value)

  def visit_Call(self, node):
    with self._scope_in():
      self.visit(node.func)
      if self._needs_hdl_call(self._pop_value()):
        self.count += 1

    for carg in node.args:
      self.visit(carg)
    for kwarg in node.keywords:
      self.visit(kwarg)


class _Frame:

  def __init__(self, fglobals, flocals, location):
    self.fglobals = fglobals
    self.flocals = flocals
    self.location = location
    self.yields = []
    self.global_names = set()
    self.return_capture = 0
    self.in_hdl = 0
    self.return_values = []
    self.retval = None
    self.loops = []
    self.loop_modes = []

  def new_locals(self, flocals):
    return pycu.new_with(self, flocals=flocals)


class _VoidResult:
  pass


class _ExecVisitor(ast.NodeVisitor):

  def __init__(self, vglobals, vlocals=None):
    super().__init__()
    self._default_loop_mode = _LoopModes.UNROLLED
    self._frames = [_Frame(vglobals, vlocals or dict(), _SourceLocation('NOFILE', 0))]
    self._variables = []
    self._results = []
    self._revgen = pycu.RevGen(fmt='{name}{ver}')

  @property
  def frame(self):
    return self._frames[-1]

  @property
  def locals(self):
    return self.frame.flocals

  @property
  def globals(self):
    return self.frame.fglobals

  @property
  def global_names(self):
    return self.frame.global_names

  @property
  def location(self):
    return self.frame.location

  @property
  def yields(self):
    return self.frame.yields

  @property
  def variables(self):
    return self._variables[-1]

  @property
  def results(self):
    return self._results[-1] if self._results else None

  @property
  def loop(self):
    return self.frame.loops[-1]

  @property
  def loop_mode(self):
    frame = self.frame
    return frame.loop_modes[-1] if frame.loop_modes else self._default_loop_mode

  def visit(self, node):
    if self.frame.in_hdl >= 0:
      method = 'visit_' + pyiu.cname(node)
      visitor = getattr(self, method, self._visit_default)
    else:
      visitor = self._static_eval

    visitor(node)

  @contextlib.contextmanager
  def _frame(self, frame):
    self._frames.append(frame)
    try:
      yield self
    finally:
      self._frames.pop()

  @contextlib.contextmanager
  def _force_hdl(self, step):
    frame = self.frame
    frame.in_hdl += step
    try:
      yield self
    finally:
      frame.in_hdl -= step

  @contextlib.contextmanager
  def _return_capture(self):
    frame = self.frame
    frame.return_capture += 1
    try:
      yield self
    finally:
      frame.return_capture -= 1

  @contextlib.contextmanager
  def _loop(self, loop):
    frame = self.frame
    frame.loops.append(loop)
    try:
      yield self
    finally:
      frame.loops.pop()

  @contextlib.contextmanager
  def _loop_mode(self, mode):
    frame = self.frame
    frame.loop_modes.append(mode)
    try:
      yield self
    finally:
      frame.loop_modes.pop()

  @contextlib.contextmanager
  def _exec_locals(self, tmp_values):
    # Override temporary values, but keep the changes to the locals.
    save_dict = dict()

    ref_dict = self.locals
    for k, v in tmp_values.items():
      save_dict[k] = ref_dict.get(k, NONE)
      ref_dict[k] = v

    try:
      yield self
    finally:
      for k, v in save_dict.items():
        if v is NONE:
          ref_dict.pop(k, None)
        else:
          ref_dict[k] = v

  @contextlib.contextmanager
  def _eval_locals(self, tmp_values):
    # Apply temporary changes during the eval operation, but revert to previous
    # status afterwards (do not persist eventual changes).
    save_dict = self.locals.copy()

    ref_dict = self.locals
    for k, v in tmp_values.items():
      ref_dict[k] = v

    try:
      yield self
    finally:
      self.locals.clear()
      self.locals.update(save_dict)

  def _store_value(self, name, value):
    if name in self.global_names:
      self.globals[name] = value
    else:
      self.locals[name] = value

  def _add_variable(self, name, dtype, isreg, init=None, vspec=None):
    alog.debug(lambda: f'NEW VAR: {valkind(isreg)} {dtype}\t{name}\t{init}\t{vspec}')

    self.variables[name] = Variable(dtype, isreg, init=init, vspec=vspec)

  def _static_eval(self, node):
    self.location.set_lineno(node.lineno)

    return asu.static_eval(node, self.globals, self.locals,
                           filename=self.location.filename)

  def _eval_node(self, node, visit_node=True):
    self.location.set_lineno(node.lineno)

    results = []
    self._results.append(results)
    try:
      if visit_node:
        self.visit(node)
      else:
        alog.debug(lambda: asu.dump(node))
        self.generic_visit(node)
    finally:
      self._results.pop()

    return results[0] if len(results) == 1 else results if results else _VoidResult()

  def _annotated_exception(self, ex):
    noted = getattr(ex, '_noted', False)
    if not noted:
      # Discard the 0-th entry as it is the NOFILE dummy one.
      locs = []
      for f in self._frames[1: ]:
        locs.append(f'{f.location.filename}:{f.location.lineno}')

      ex = pycu.rewrited_exception(ex, f'\nError stack:\n' + '\n'.join(locs))
      setattr(ex, '_noted', True)

    return ex

  def eval_node(self, node, visit_node=True):
    try:
      return self._eval_node(node, visit_node=visit_node)
    except _Exception:
      # These are internal exceptions meant to be left as is.
      raise
    except Exception as ex:
      raise self._annotated_exception(ex) from None

  def push_result(self, result):
    results = self.results
    if results is not None and not isinstance(result, _VoidResult):
      results.append(result)

  def push_yield(self, value):
    yields = self.yields
    if yields is not None:
      yields.append(value)

  def _get_function_body(self, name, node):
    if isinstance(node, ast.Module):
      for mnode in node.body:
        if isinstance(mnode, ast.FunctionDef) and mnode.name == name:
          return mnode.body

    alog.warning(f'Function "{name}" body not found in node: {asu.dump(node)}')

    return [node]

  def _populate_args_locals(self, sig, args, kwargs, func_locals):
    alog.debug(lambda: f'Build Args: ARGS={args}\tKWARGS={kwargs}\tFNLOCALS={func_locals}')

    n, xkwargs = 0, kwargs.copy()
    for param in sig.parameters.values():
      if param.kind == inspect.Parameter.POSITIONAL_ONLY:
        func_locals[param.name] = args[n]
        n += 1
      elif param.kind == inspect.Parameter.VAR_POSITIONAL:
        func_locals[param.name] = args[n: ]
        n = len(args)
      elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
        pvalue = xkwargs.pop(param.name, NONE)
        if pvalue is not NONE:
          func_locals[param.name] = pvalue
        elif param.default != inspect.Parameter.empty:
          func_locals[param.name] = param.default
        else:
          func_locals[param.name] = args[n]
          n += 1
      elif param.kind == inspect.Parameter.KEYWORD_ONLY:
        pvalue = xkwargs.pop(param.name, NONE)
        if pvalue is NONE and param.default != inspect.Parameter.empty:
          pvalue = param.default

        func_locals[param.name] = pvalue
      elif param.kind == inspect.Parameter.VAR_KEYWORD:
        func_locals[param.name] = xkwargs
        xkwargs = dict()

    return xkwargs

  def _generate_retval_result(self, fname, value, vindex, tmp_names):
    if isinstance(value, (list, tuple)):
      results = []
      for lvalue in value:
        rvalue, vindex = self._generate_retval_result(fname, lvalue, vindex, tmp_names)
        results.append(rvalue)

      return type(value)(results), vindex
    elif pycu.isdict(value):
      rdict = dict()
      for k, dvalue in value.items():
        rvalue, vindex = self._generate_retval_result(fname, dvalue, vindex, tmp_names)
        rdict[k] = rvalue

      return rdict, vindex

    if vindex == len(tmp_names):
      rvname = self._revgen.newname(fname)
      tmp_names.append(rvname)
    elif vindex < len(tmp_names):
      rvname = tmp_names[vindex]
    else:
      fatal(f'Value index out of range: {vindex} vs {len(tmp_names)}')

    var = self.load_var(rvname, ctx=ast.Store())
    if var is NONE and isinstance(value, Value):
      var = self._new_variable(rvname, value)

    self._assign_value(var, value, rvname)

    return self.load_var(rvname, ctx=ast.Load()), vindex + 1

  def _generate_retlist_result(self, fname, return_values):
    tmp_names, results, sig = [], [], None
    for i, retval in enumerate(return_values):
      with self.emitter.placement(retval.placement):
        vsig = pycu.signature(retval.value)
        if sig is not None and not pycu.equal_signature(sig, vsig):
          fatal(f'Return values signature mismatch: {vsig} vs. {sig}')

        sig = vsig

        rvalue, _ = self._generate_retval_result(fname, retval.value, 0, tmp_names)
        results.append(rvalue)

    return results[0]

  def _run_function_helper(self, func, args, kwargs):
    func_self = getattr(func, '__self__', None)
    func = getattr(func, '__wrapped__', func)

    sig = inspect.signature(func)
    alog.debug(lambda: f'Signature: {sig}')

    fninfo = get_function_info(func)
    alog.debug(lambda: f'Source: {fninfo.filename} @ {fninfo.lineno}\n{fninfo.source}')

    func_node = ast.parse(fninfo.source, filename=fninfo.filename, mode='exec')

    func_node = ast_hdl_transform(func_node)
    alog.debug(lambda: f'FUNC AST: {asu.dump(func_node)}')

    func_locals = dict()
    if func_self is not None:
      args = [func_self] + args

    kwargs = self._populate_args_locals(sig, args, kwargs, func_locals)
    func_locals.update(**kwargs)

    func_body = self._get_function_body(pyiu.func_name(func), func_node)

    frame = _Frame(get_obj_globals(func),
                   func_locals,
                   _SourceLocation(fninfo.filename, fninfo.lineno))
    with self._frame(frame):
      try:
        for bnode in func_body:
          self.visit(bnode)
      except _ReturnException as rex:
        frame.retval = rex.value

    if frame.return_values:
      # This handles the case of return statements from within an HDL function.
      # As all the HDL functions gets inlined, there is not really a "return", so
      # the returned values get stored into a temporary, and a "load" of that temporary
      # returned from this API.
      result = self._generate_retlist_result(pyiu.func_name(func), frame.return_values)
    else:
      result = frame.yields if frame.yields else frame.retval

    alog.debug(lambda: f'RESULT: {result}\tEXIT LOCALS: {pyu.stri(func_locals)}')

    return result

  def _run_class_function(self, func, args, kwargs):
    alog.debug(lambda: f'OBJECT CREATE: class={pyiu.func_name(func)} args={pyu.stri(args)} ' \
               f'kwargs={pyu.stri(kwargs)}')
    obj = func.__new__(func, *args, **kwargs)
    init = getattr(func, '__init__', None)
    if init is not None:
      self._run_function_helper(init, [obj] + args, kwargs)

    return obj

  def _call_direct(self, func, args, kwargs):
    alog.debug(lambda: f'DIRECT CALL: function={pyiu.func_name(func)} args={pyu.stri(args)} ' \
               f'kwargs={pyu.stri(kwargs)}')

    # We cannot call func(*args, **kwargs) directly as we need to insert the current
    # locals and globals.
    self.locals.update(__func=func, __args=args, __kwargs=kwargs)

    return eval('__func(*__args, **__kwargs)', self.globals, self.locals)

  def _needs_hdl_processing(self, func):
    if is_hdl_function(func):
      return True
    elif not inspect.isclass(func):
      return False

    return is_hdl_function(getattr(func, '__init__', None))

  def run_function(self, func, args, kwargs=None):
    kwargs = kwargs or dict()
    if self._needs_hdl_processing(func):
      if inspect.isclass(func):
        # Running a function with a class function object, means object creation.
        result = self._run_class_function(func, args, kwargs)
      else:
        result = self._run_function_helper(func, args, kwargs)

      return result

    return self._call_direct(func, args, kwargs)

  def _is_hdl_tree(self, node):
    checker = _HdlChecker(self.globals, self.locals)
    checker.visit(node)

    return checker.count > 0


class CodeGen(_ExecVisitor):

  def __init__(self, emitter, globs):
    super().__init__(globs)
    self.emitter = emitter
    self._module_decls_place = emitter.emit_placement()
    self._used_entities = set()
    self._generated_entities = set()
    self._ent_versions = EntityVersions()
    self._vars_places = []
    self._root_vars = dict()
    self._process_kind = None
    self._process_args = None
    self._setup_handlers()

  def _setup_handlers(self):
    self.visit_IfExp = functools.partial(self._hdl_visitor, 'IfExp', pushres=True)
    self.visit_If = functools.partial(self._hdl_visitor, 'If')
    self.visit_For = functools.partial(self._hdl_visitor, 'For')
    self.visit_While = functools.partial(self._hdl_visitor, 'While')
    self.visit_Match = functools.partial(self._hdl_visitor, 'Match')
    self.visit_Try = functools.partial(self._hdl_visitor, 'Try')
    self.visit_With = functools.partial(self._hdl_visitor, 'With')
    self.visit_Call = functools.partial(self._hdl_visitor, 'Call', pushres=True)

  def _hdl_visitor(self, name, node, pushres=False):
    if self.frame.in_hdl > 0 or self._is_hdl_tree(node):
      handler = getattr(self, f'_handle_{name}')
      handler(node)
    else:
      result = self._static_eval(node)
      if pushres:
        self.push_result(result)

  def _get_format_parts(self, fmt, kwargs):

    def mapfn(tok):
      arg = self.emitter.eval_token(tok)
      if arg is None:
        varg = self.run_code(tok, kwargs, 'eval',
                             filename=self.location.filename,
                             lineno=self.location.lineno)
        return self.emitter.eval_tostring(varg)

      return arg

    def nmapfn(tok):
      return self.emitter.quote_string(tok)

    return pyu.sreplace(_CODEFMT_RX, fmt, mapfn, nmapfn=nmapfn, join=False)

  def _resolve_code(self, code, kwargs):

    def mapfn(tok):
      arg = self.emitter.eval_token(tok)
      if arg is None:
        varg = self.run_code(tok, kwargs, 'eval',
                             filename=self.location.filename,
                             lineno=self.location.lineno)
        return self.emitter.svalue(varg)

      return arg

    return pyu.sreplace(_CODEFMT_RX, code, mapfn)

  def _visit_default(self, node):
    # This is the default handler when a specific visit_XXX() method does not exist.
    # If something is not working, look for 'FAST STATIC' within the logs, which will
    # tell which node is escaping from the PyXHDL interpreter.
    alog.debug(lambda: f'FAST STATIC: {asu.dump(node)}')
    self._static_eval(node)

  def load_var(self, name, ctx=ast.Load()):
    var = vload(name, self.globals, self.locals)
    if var is NONE:
      shvar = self._root_vars.get(name)
      if shvar is not None:
        var = Value(shvar.dtype, Ref(name, vspec=shvar.vspec, vname=name),
                    isreg=shvar.isreg)
      elif isinstance(ctx, ast.Load):
        fatal(f'Undefined variable: {name}')

    return self.emitter.var_remap(var, isinstance(ctx, ast.Store))

  def _new_variable(self, name, value):
    isreg = value.isreg if value.isreg is not None else False
    base_name = None

    vinit = value.init
    if vinit is not None:
      init, vspec, base_name = vinit.value, vinit.vspec, vinit.name
    else:
      init, vspec = None, None

      vref = value.ref
      if vref is not None:
        vspec = vref.vspec
        if vref.cname is not None:
          base_name = vref.cname

    if vspec is not None:
      vspec = vspec.for_new_variable()

    vname = self._revgen.newname(base_name or name, shortzero=True)

    var = Value(value.dtype, Ref(vname, vspec=vspec, vname=vname), isreg=isreg)
    self._add_variable(vname, var.dtype, var.isreg, init=init, vspec=vspec)
    self._store_value(name, var)

    return var

  def _assign_variable(self, var, value, name):
    if isinstance(value, Value) and isinstance(value.value, Init):
      if value.isreg != var.isreg:
        alog.warning(f'Cannot create "{var.name}" as {valkind(value.isreg)}, ' \
                     f'will be {valkind(var.isreg)}')
      if isinstance(value.value, Init):
        alog.debug(lambda: f'ASSIGN CREATE: {name} is {value.dtype} = {value.value}')
      else:
        alog.debug(lambda: f'ASSIGN CREATE: {name} is {value.dtype}')
    else:
      if is_ro_ref(var):
        fatal(f'Trying to assign {var.name} which is read-only')

      self.emitter.emit_assign(var, name, value)

  def _assign_value(self, var, value, name):
    alog.debug(lambda: f'ASSIGN: {var} ({name}) = {value}')

    if not isinstance(var, Value):
      if isinstance(value, Value):
        # Two cases of creating new variables.
        if value.init is not None:
          # The mkwire(name=None), mkreg(name=None), mkvwire() and mkvreg() APIs
          # have the "init" member populated.
          var = self._new_variable(name, value)
        elif (ref := value.ref) and ref.cname is not None:
          # The mkwire(name=XYZ) and mkreg(name=XYZ) APIs have the "ref" populated,
          # with ref.cname not None.
          var = self._new_variable(name, value)
          return
    elif name is not None and var.ref is None:
      var = None

    if isinstance(var, Value):
      self._assign_variable(var, value, name)
    elif name is not None:
      self._store_value(name, value)

  def _multi_assign_inner(self, target, value):
    elts = elements(target)
    if elts is not None:
      for el, val in zip(elts, value):
        self._multi_assign_inner(el, val)
    else:
      var = self.eval_node(target)
      if isinstance(var, _Storer):
        var.store(value)
      else:
        name = target.id if isinstance(target, ast.Name) else None
        self._assign_value(var, value, name)

  def _multi_assign(self, targets, value):
    for target in targets:
      self._multi_assign_inner(target, value)

  def assign_value(self, name, value):
    self._assign_value(self.load_var(name, ctx=ast.Store()), value, name)

  def _unpack_value(self, targets, values, dest):
    if isinstance(targets, ast.Name):
      dest[targets.id] = values
    else:
      elts = elements(targets)
      if elts:
        for t, v in zip(elts, values):
          self._unpack_value(t, v, dest)

  def _handle_array(self, node, atype):
    alog.debug(lambda: asu.dump(node))
    values = [self.eval_node(n) for n in node.elts]

    self.push_result(atype(values))

  def _handle_call(self, func, args, kwargs):
    if pyiu.is_subclass(func, Entity):
      alog.debug(lambda: f'Entity instantiation: {pyiu.func_name(func)} args={pyu.stri(args)} ' \
                 f'kwargs={pyu.stri(kwargs)}')
      # We do not run the Entity init code with run_function() as this is not what
      # we want. Doing so will end up emitting HDL code.
      # Entities constructors cannot perform operations which lead to HDL code emission.
      # Note that HDL data type manipulation (like adding values to containers and such)
      # is OK, as this will not emit HDL code.
      result = func(*args, **kwargs)

      ent_name = self._register_entity(func, kwargs)

      self.emitter.emit_entity(result, kwargs, ent_name=ent_name)
    else:
      result = self.run_function(func, args, kwargs=kwargs)

    return result

  def _process_scope_enter(self, vars_scope, process_kind, process_args):
    self._variables.append(dict())
    self._vars_places.append(vars_scope)
    self._process_kind = process_kind
    self._process_args = process_args

  def _process_scope_exit(self):
    alog.debug(lambda: f'Variables stack is {len(self._variables)} deep')

    svars = self._variables.pop()
    place = self._vars_places.pop()

    root_vars = dict()
    with self.emitter.placement(place) as emt:
      for name, var in svars.items():
        if emt.is_root_variable(var):
          root_vars[name] = var
        else:
          alog.debug(lambda: f'VARIABLE: {valkind(var.isreg)} {var.dtype} {name}')
          emt.emit_declare_variable(name, var)

    with self.emitter.placement(self.emitter.module_vars_place) as emt:
      for name, var in root_vars.items():
        shv = self._root_vars.get(name)
        if shv is not None:
          if shv != var:
            fatal(f'Root variable declaration mismatch: {pyu.stri(var)} vs. {pyu.stri(shv)}')
        else:
          alog.debug(lambda: f'ROOT VARIABLE: {valkind(var.isreg)} {var.dtype} {name}')
          emt.emit_declare_variable(name, var)

    self._root_vars.update(**root_vars)

  def _register_entity(self, eclass, kwargs, generated=False):
    pargs, rkwargs = dict(), kwargs.copy()
    for pin in eclass.PORTS:
      arg = rkwargs.pop(pin.name, None)
      if arg is None:
        fatal(f'Missing entity port "{pin.name}" binding for entity {eclass.__name__}')

      pin.verify_arg(arg)
      if pin.is_ifc():
        pargs[pin.name] = pin.ifc_view(arg)
      else:
        pargs[pin.name] = arg.new_value(make_port_ref(pin))

    for kwarg_name, arg in eclass.ARGS.items():
      rkwargs[kwarg_name] = rkwargs.get(kwarg_name, arg)

    ename = eclass.NAME
    if ename is None:
      # External entities (for which the NAME field is provided within the Entity
      # subclass declaration) do not need to be regsitered, as no generation needs
      # to happen for them.
      ename, erec = self._ent_versions.getname(eclass, pargs, {k: wrap(v) for k, v in rkwargs.items()})
      self._used_entities.add(erec)
      if generated:
        self._generated_entities.add(erec)
    else:
      # External libraries can provide a specific library to be included (where
      # such entity is defined).
      if eclass.LIBNAME:
        self.emitter.add_extra_library(eclass.LIBNAME)

    return ename

  def _get_sensitivity(self, hdl_args, din):
    def expand(v, dest):
      if pycu.isdict(v):
        dest.update(v)
      elif isinstance(v, str):
        for port in pyu.resplit(v, ','):
          sargs = dict()
          if port.startswith('+'):
            sargs['trigger'] = POSEDGE
            port = port[1: ]
          elif port.startswith('-'):
            sargs['trigger'] = NEGEDGE
            port = port[1: ]

          dest[port] = Sens(**sargs)
      elif isinstance(v, (list, tuple)):
        for x in v:
          expand(x, dest)

      return dest

    sstmp = expand(hdl_args.get('sens', dict()), dict())

    sensitivity = dict()
    for name, sens in sstmp.items():
      m = re.match(r'(\w+)(\.(\w+))?', name)
      if not m:
        fatal(f'Invalid sensitivity port name: {name}')

      pname, fname = m.group(1), m.group(3)

      if fname:
        pin = din.get(pname)
        if pin is None:
          fatal(f'Sensitivity source interface is not a port: {pname}')
        if not pin.is_ifc():
          fatal(f'Sensitivity source interface is not an interface: {pin}')

        pname = pin.ifc_field_name(fname)
      else:
        if pname not in din and pname not in self._root_vars:
          fatal(f'Sensitivity source is not a port: {pname}')

      sensitivity[pname] = sens

      alog.debug(lambda: f'Sensitivity: {pname} {sens.trigger}')

    return sensitivity

  def generate_entity(self, eclass, eargs):
    alog.debug(lambda: f'Entity {eclass.__name__}')

    ent_name = self._register_entity(eclass, eargs, generated=True)

    kwargs = eargs.copy()
    din, cargs = dict(), dict()
    for pin in eclass.PORTS:
      alog.debug(lambda: f'Port: {pin.name} {pin.idir}')

      arg = kwargs.pop(pin.name, None)
      if arg is None:
        fatal(f'Missing argument "{pin.name}" for Entity "{eclass.__name__}"')

      din[pin.name] = pin
      if pin.is_ifc():
        cargs[pin.name] = pin.ifc_view(arg)
      else:
        ref = make_port_ref(pin)
        if isinstance(arg, Value):
          port_arg = arg.new_value(ref)
        else:
          if not isinstance(arg, Type):
            fatal(f'Argument must be Type at this point: {arg}')

          port_arg = mkwire(arg, name=ref)

        cargs[pin.name] = self.emitter.make_port_arg(port_arg)

    for kwarg_name, arg in eclass.ARGS.items():
      rarg = kwargs.get(kwarg_name, arg)
      kwargs[kwarg_name] = rarg

      alog.debug(lambda: f'Arg: {kwarg_name} = {rarg}')

    uw_kwargs = {k: unwrap(v) for k, v in kwargs.items()}
    ent_args = pycu.dmerge(cargs, uw_kwargs)

    ent = eclass(**ent_args)

    ecomm = f'Entity "{ent_name}" is "{eclass.__name__}" with:\n' \
      f'\targs={({k: pyu.stri(v.dtype if isinstance(v, Value) else v) for k, v in cargs.items()})}\n' \
      f'\tkwargs={pyu.stri(ent.kwargs)}'

    with self.emitter.placement(self._module_decls_place) as emt:
      emt.emit_module_def(ent_name, ent, comment=ecomm)

    self.emitter.emit_module_decl(ent_name, ent)

    for func in ent.enum_processes():
      hdl_args = get_hdl_args(func) or dict()

      alog.debug(lambda: f'Process function: {pyiu.func_name(func)}')

      sensitivity = self._get_sensitivity(hdl_args, din)
      process_kind = hdl_args.get('kind')

      with self.emitter.indent():
        # Process functions will automatically see port names as locals, so the
        # position arguments list is empty. Process functions can still have keyword
        # arguments, which will be correctly populated.
        self.generate_process(func, [ent],
                              kwargs=ent_args,
                              sensitivity=sensitivity,
                              process_kind=process_kind,
                              process_args=hdl_args)

    self.emitter.emit_module_end()

    self._reset_entity_context()

  def generate_process(self, func, args, kwargs=None, sensitivity=None,
                       process_kind=None, process_args=None):
    if process_kind == ROOT_PROCESS:
      self._process_scope_enter(self.emitter.module_vars_place, process_kind,
                                process_args)
      try:
        result = self.run_function(func, args, kwargs=kwargs)
      finally:
        self._process_scope_exit()
    else:
      self.emitter.emit_process_decl(pyiu.func_name(func),
                                     sensitivity=sensitivity,
                                     process_kind=process_kind,
                                     process_args=process_args)

      self.emitter.emit_process_begin()
      self._process_scope_enter(self.emitter.process_vars_place, process_kind,
                                process_args)
      try:
        with self.emitter.indent():
          result = self.run_function(func, args, kwargs=kwargs)
      finally:
        self._process_scope_exit()
      self.emitter.emit_process_end()

    return result

  def _reset_entity_context(self):
    self._root_vars = dict()
    self._revgen = pycu.RevGen(fmt='{name}{ver}')

  def _flush_generation(self):
    while True:
      used_entities = self._used_entities.copy()
      generated = 0
      for erec in used_entities:
        if erec not in self._generated_entities:
          self.generate_entity(erec.eclass, pycu.dmerge(erec.pargs, erec.kwargs))
          generated += 1

      if generated == 0:
        break

  def visit_Constant(self, node):
    self.push_result(node.value)

  def visit_FormattedValue(self, node):
    self.push_result(self._static_eval(node))

  def visit_JoinedStr(self, node):
    self.push_result(self._static_eval(node))

  def visit_Tuple(self, node):
    self._handle_array(node, tuple)

  def visit_List(self, node):
    self._handle_array(node, list)

  def visit_Set(self, node):
    alog.debug(lambda: asu.dump(node))
    result = set()
    for snode in node.elts:
      result.add(self.eval_node(snode))

    self.push_result(result)

  def visit_Dict(self, node):
    alog.debug(lambda: asu.dump(node))
    result = dict()
    for knode, vnode in zip(node.keys, node.values):
      k = self.eval_node(knode)
      v = self.eval_node(vnode)
      result[k] = v

    self.push_result(result)

  def visit_Name(self, node):
    alog.debug(lambda: asu.dump(node))
    result = self.load_var(node.id, ctx=node.ctx)
    self.push_result(result)

  def visit_Expression(self, node):
    value = self.eval_node(node.body)
    self.push_result(value)

  def visit_Expr(self, node):
    value = self.eval_node(node.value)
    self.push_result(value)

  def visit_UnaryOp(self, node):
    operand = self.eval_node(node.operand)
    if has_hdl_vars(operand):
      result = self.emitter.eval_UnaryOp(node.op, operand)
    else:
      result = self._static_eval(node)

    self.push_result(result)

  def _neval(self, ref_node, cls, **kwargs):
    snode = cls(**kwargs)
    ast.copy_location(snode, ref_node)

    return self._static_eval(snode)

  def visit_BinOp(self, node):
    left = self.eval_node(node.left)
    right = self.eval_node(node.right)
    if has_hdl_vars((left, right)):
      result = self.emitter.eval_BinOp(node.op, left, right)
    else:
      result = self._static_eval(node)

    self.push_result(result)

  def visit_BoolOp(self, node):
    alog.debug(lambda: asu.dump(node))

    # Implement AND(False, ...) and OR(True, ...) shortcutting.
    result, values = None, []
    for val in node.values:
      xval = self.eval_node(val)
      if xval is True:
        if isinstance(node.op, ast.Or):
          result = True
          break
      elif xval is False:
        if isinstance(node.op, ast.And):
          result = False
          break
      else:
        values.append(xval)

    if result is None:
      if has_hdl_vars(values):
        result = self.emitter.eval_BoolOp(node.op, values)
      else:
        result = self._static_eval(node)

    self.push_result(result)

  def visit_Compare(self, node):
    left = self.eval_node(node.left)
    comparators = []
    for comp in node.comparators:
      xcomp = self.eval_node(comp)
      comparators.append(xcomp)

    if has_hdl_vars((left, comparators)):
      result = self.emitter.eval_Compare(left, node.ops, comparators)
    else:
      result = self._static_eval(node)

    self.push_result(result)

  def visit_Lambda(self, node):
    alog.debug(lambda: asu.dump(node))

    flocals = self.locals

    def lambda_runner(*args, **kwargs):
      func_locals = flocals.copy()

      for p, arg in zip(node.args.args, args):
        func_locals[p.arg] = arg

      func_locals.update(**kwargs)

      with self._frame(self.frame.new_locals(func_locals)):
        result = self.eval_node(node.body)

      alog.debug(lambda: f'LM RESULT {result}\tLOCALS: {pyu.stri(func_locals)}')

      return result

    self.push_result(lambda_runner)

  def _handle_Call(self, node):
    func = self.eval_node(node.func)

    args = []
    for carg in node.args:
      if isinstance(carg, ast.Starred):
        value = self.eval_node(carg.value)
        for svalue in elements(value):
          args.append(svalue)
      else:
        value = self.eval_node(carg)
        args.append(value)

    kwargs = dict()
    for kwarg in node.keywords:
      kwval = self.eval_node(kwarg.value)
      if kwarg.arg:
        kwargs[kwarg.arg] = kwval
      else:
        if not pycu.isdict(kwval):
          fatal(f'Wrong type: {type(kwval)}')

        for name, value in kwval.items():
          kwargs[name] = value

    alog.debug(lambda: f'CALL: {func}\t{pyu.stri(args)}\t{pyu.stri(kwargs)}')

    result = self._handle_call(func, args, kwargs)

    self.push_result(result)

  def _load_subs_attr(self, node, ctx):
    if isinstance(node, ast.Name):
      return self.load_var(node.id, ctx=ctx)

    return self.eval_node(node)

  def _attr_storefn(self, target, name):

    def setfn(value):
      var = getattr(target, name, None)
      if isinstance(var, Value) and var.ref is not None:
        self._assign_value(var, value, name)
      else:
        setattr(target, name, value)

    return setfn

  def visit_Attribute(self, node):
    value = self._load_subs_attr(node.value, node.ctx)
    if isinstance(node.ctx, ast.Load):
      result = getattr(value, node.attr)
    elif isinstance(node.ctx, ast.Store):
      if isinstance(value, Value) and is_ro_ref(value):
        fatal(f'Trying to assign attribute {node.attr} of {value.name}, which is read-only')

      result = _Storer(self._attr_storefn(value, node.attr))
    else:
      fatal(f'Unknown subscript context type: {asu.dump(node.ctx)}')

    self.push_result(result)

  def visit_NamedExpr(self, node):
    value = self.eval_node(node.value)
    var = self.load_var(node.target.id, ctx=node.target.ctx)
    self._assign_value(var, value, node.target.id)

  def visit_Subscript(self, node):
    value = self._load_subs_attr(node.value, node.ctx)
    sv = self.eval_node(node.slice)

    if isinstance(value, Value) and isinstance(node.ctx, ast.Store) and is_ro_ref(value):
      fatal(f'Trying to assign subscript {sv} of {value.name}, which is read-only')

    if isinstance(value, Value):
      result = self.emitter.eval_Subscript(value, sv)
    elif isinstance(node.ctx, ast.Load):
      result = value[sv]
    elif isinstance(node.ctx, ast.Store):
      result = _Storer(subscript_setter(value, sv))
    else:
      fatal(f'Unknown subscript context type: {asu.dump(node.ctx)}')

    self.push_result(result)

  def visit_Slice(self, node):
    lower = self.eval_node(node.lower) if node.lower is not None else None
    upper = self.eval_node(node.upper) if node.upper is not None else None
    step = self.eval_node(node.step) if node.step is not None else None
    if has_hdl_vars((upper, step)):
      fatal(f'Slice cannot have HDL arguments: {asu.dump(node)}')

    self.push_result(slice(lower, upper, step))

  def _walk_generator(self, node, fn):
    for gen in node.generators:
      gen_iter = self.eval_node(gen.iter)
      for xv in gen_iter:
        self.locals[gen.target.id] = xv

        ifv = True
        for ifn in gen.ifs:
          ifv = self.eval_node(ifn)
          if not ifv:
            break

        if ifv:
          fn(node)

  def visit_DictComp(self, node):
    result = dict()

    def walker(node):
      kv = self.eval_node(node.key)
      vv = self.eval_node(node.value)
      result[kv] = vv

    self._walk_generator(node, walker)

    self.push_result(result)

  def _handle_container_comp(self, node, ctype):
    result = []

    def walker(node):
      cv = self.eval_node(node.elt)
      result.append(cv)

    self._walk_generator(node, walker)

    if not isinstance(result, ctype):
      result = ctype(result)

    self.push_result(result)

  def visit_ListComp(self, node):
    self._handle_container_comp(node, list)

  def visit_SetComp(self, node):
    self._handle_container_comp(node, set)

  def visit_GeneratorExp(self, node):
    self._handle_container_comp(node, tuple)

  def visit_Assign(self, node):
    value = self.eval_node(node.value)
    self._multi_assign(node.targets, value)

  def visit_AugAssign(self, node):
    ltarget = as_loading(node.target)
    ref_value = self.eval_node(ltarget)
    value = self.eval_node(node.value)

    if has_hdl_vars((ref_value, value)):
      result = self.emitter.eval_BinOp(node.op, ref_value, value)
    else:
      result = self._neval(node, ast.BinOp, op=node.op, left=ltarget, right=node.value)

    self._multi_assign([node.target], result)

  def visit_Raise(self, node):
    exc = self.eval_node(node.exc)
    raise exc

  def visit_Assert(self, node):
    test = self.eval_node(node.test)
    msg = self.eval_node(node.msg) if node.msg is not None else None

    if has_hdl_vars(test):
      parts = self._get_format_parts(msg, dict()) if msg else None
      self.emitter.emit_Assert(test, parts)
    elif not test:
      fatal(msg, exc=AssertionError)

  def visit_Module(self, node):
    # This handler simply dive into inner nodes by calling the AST Visitor
    # generic_visit() API (after some logging we do to make sense of what is
    # being processed).
    self.eval_node(node, visit_node=False)

  def _handle_IfExp(self, node):
    test = self.eval_node(node.test)
    if has_hdl_vars(test):
      body = self.eval_node(node.body)
      orelse = self.eval_node(node.orelse)
      result = self.emitter.eval_IfExp(test, body, orelse)
    else:
      if test:
        result = self.eval_node(node.body)
      else:
        result = self.eval_node(node.orelse)

    self.push_result(result)

  def _handle_If(self, node):
    test = self.eval_node(node.test)

    if has_hdl_vars(test):
      with self._return_capture():
        self.emitter.emit_If(test)
        with self.emitter.indent():
          for insn in node.body:
            self.eval_node(insn)

        enode, enodes = node, []
        while len(enode.orelse) == 1 and isinstance(enode.orelse[0], ast.If):
          enode = enode.orelse[0]
          etest = self.eval_node(enode.test)
          if has_hdl_vars(etest):
            self.emitter.emit_Elif(etest)
            with self.emitter.indent():
              for insn in enode.body:
                self.eval_node(insn)
          elif etest:
            # Instructions which are part of a statically (no HDL) True "elif" branch
            # are emitted after the "if" instruction.
            enodes.extend(enode.body)

        if enode.orelse or enodes:
          self.emitter.emit_Else()
          with self.emitter.indent():
            for insn in enodes + (enode.orelse or []):
              self.eval_node(insn)

        self.emitter.emit_EndIf()
    else:
      alog.debug(lambda: f'Resolving static If test: {asu.dump(node.test)}')
      if test:
        for insn in node.body:
          self.eval_node(insn)
      else:
        for insn in node.orelse:
          self.eval_node(insn)

  def _decode_for_loop(self, node):
    idata = self.eval_node(node.iter)

    if isinstance(node.target, ast.Name):
      idata = tuple(idata)
      if idata and all(isinstance(x, int) for x in idata):
        steps = set(idata[i + 1] - idata[i] for i in range(len(idata) - 1))
        if len(steps) <= 1:
          return _ForLoop(data=idata,
                          ivar=node.target.id,
                          start=idata[0],
                          end=idata[-1],
                          step=steps.pop() if steps else 1)

    return _ForLoop(data=idata)

  def _unrolled_For(self, node, floop):
    with self._loop(_LoopContext(mode=_LoopModes.UNROLLED)):
      for t in floop.data:
        self._unpack_value(node.target, t, self.locals)

        got_break = False
        for insn in node.body:
          try:
            self.eval_node(insn)
          except _BreakException:
            got_break = True
            break
          except _ContinueException:
            break

        if got_break:
          break

  def _hdl_For(self, node, floop):
    loop_var = mkwire(INT, name=floop.ivar, const=True)

    with (self._loop(_LoopContext(mode=_LoopModes.HDL)),
          self._exec_locals({floop.ivar: loop_var})):
      self.emitter.emit_For(floop.ivar, floop.start, floop.end, floop.step)
      with self.emitter.indent():
        for insn in node.body:
          self.eval_node(insn)

      self.emitter.emit_EndFor()

  def _handle_For(self, node):
    alog.debug(lambda: asu.dump(node))

    floop = self._decode_for_loop(node)

    if floop.ivar is None or self.loop_mode != _LoopModes.HDL:
      self._unrolled_For(node, floop)
    else:
      self._hdl_For(node, floop)

  def _handle_While(self, node):
    alog.debug(lambda: asu.dump(node))

    with self._loop(_LoopContext(mode=_LoopModes.UNROLLED)):
      while True:
        test = self.eval_node(node.test)
        if has_hdl_vars(test):
          fatal(f'While test cannot have HDL vars: {asu.dump(node.test)}')

        if test:
          got_break = False
          for insn in node.body:
            try:
              self.eval_node(insn)
            except _BreakException:
              got_break = True
              break
            except _ContinueException:
              break

          if got_break:
            break
        else:
          break

  def visit_Break(self, node):
    alog.debug(lambda: asu.dump(node))

    loop = self.loop
    if loop.mode == _LoopModes.UNROLLED:
      raise _BreakException()
    else:
      self.emitter.emit_Break()

  def visit_Continue(self, node):
    alog.debug(lambda: asu.dump(node))

    loop = self.loop
    if loop.mode == _LoopModes.UNROLLED:
      raise _ContinueException()
    else:
      self.emitter.emit_Continue()

  def _handle_Try(self, node):
    try:
      for bnode in node.body:
        self.visit(bnode)
    except Exception as e:
      alog.debug(lambda: f'Caught exception: {e}')
      for xhand in node.handlers:
        xtype = self.eval_node(xhand.type)
        if isinstance(e, xtype):
          if xhand.name:
            self.locals[xhand.name] = e
          for bnode in xhand.body:
            self.visit(bnode)

          break
    else:
      for enode in node.orelse:
        self.visit(enode)
    finally:
      for fnode in node.finalbody:
        self.visit(fnode)

  def _handle_With(self, node):
    items, names = [], []
    for inode in node.items:
      item = self.eval_node(inode.context_expr)
      name = inode.optional_vars.id if isinstance(inode.optional_vars, ast.Name) else None
      items.append(item)
      names.append(name)

    bi, bitems, rex = 0, [], None
    try:
      while bi < len(items):
        bitem = items[bi].__enter__()
        bitems.append(bitem)
        bi += 1

      for name, bitem in zip(names, bitems):
        if name is not None:
          self.locals[name] = bitem

      for bnode in node.body:
        self.eval_node(bnode)

    except Exception as ex:
      rex = ex

    wex = None
    bi -= 1
    while bi >= 0:
      if rex is not None:
        if not items[bi].__exit__(type(rex), rex, rex.__traceback__):
          wex = rex
      else:
        items[bi].__exit__(None, None, None)
      bi -= 1

    if wex is not None:
      raise wex

  def visit_Return(self, node):
    alog.debug(lambda: asu.dump(node))
    value = self.eval_node(node.value) if node.value is not None else None
    if self.frame.return_capture:
      retval = _Return(value=value, placement=self.emitter.emit_placement())
      self.frame.return_values.append(retval)
    else:
      raise _ReturnException(value)

  def visit_Yield(self, node):
    # Yielded values are accumulated into yields list setup by the function call
    # processing. This is different from how they are implemented in CPython but
    # they are much easier to implement.
    alog.debug(lambda: asu.dump(node))
    value = self.eval_node(node.value)
    self.push_yield(value)

  def visit_YieldFrom(self, node):
    # Yielded values are accumulated into yields list setup by the function call
    # processing. This is different from how they are implemented in CPython but
    # they are much easier to implement.
    alog.debug(lambda: asu.dump(node))
    yiter = self.eval_node(node.value)
    for value in yiter:
      self.push_yield(value)

  def visit_Global(self, node):
    for name in node.names:
      self.global_names.add(name)

  def visit_Pass(self, node):
    pass

  def _handle_Match(self, node):
    subject = self.eval_node(node.subject)
    cases = []
    with self._return_capture():
      for mc in node.cases:
        pattern = self.eval_node(mc.pattern)
        scope = self.emitter.create_placement(extra_indent=2)
        with self.emitter.placement(scope):
          for insn in mc.body:
            self.eval_node(insn)

        for ptrn in pyu.as_sequence(pattern, t=(tuple, list)):
          cases.append(_MatchCase(pattern=ptrn, scope=scope))

    self.emitter.emit_match_cases(subject, cases)

  def visit_MatchAs(self, node):
    alog.debug(lambda: asu.dump(node))
    if node.pattern:
      result = self.eval_node(node.pattern)
      assert node.name is None, 'TBH'
    elif node.name:
      result = self.load_var(node.name, ctx=ast.Load())
    else:
      # None match means "everything else".
      result =  None

    self.push_result(result)

  def visit_MatchValue(self, node):
    alog.debug(lambda: asu.dump(node))
    result = self.eval_node(node.value)
    self.push_result(result)

  def visit_MatchSequence(self, node):
    alog.debug(lambda: asu.dump(node))
    patterns = []
    for ptrn in node.patterns:
      value = self.eval_node(ptrn)
      patterns.append(value)

    self.push_result(patterns)

  @staticmethod
  def current():
    return pytc.get_context(_CGENCTX)

  def context(self):
    return pytc.Context(_CGENCTX, self)

  def run_code(self, code, args, mode, filename='<string>', lineno=1):
    dcode = textwrap.dedent(code)

    try:
      cnode = ast.parse(dcode, filename=filename, mode=mode)
    except Exception as ex:
      raise pycu.rewrited_exception(ex, f'\nCode:\n{dcode}') from None

    cnode.lineno = lineno
    ast.fix_missing_locations(cnode)

    alog.debug(lambda: f'RUN CODE: {asu.dump(cnode)}')

    if mode == 'exec':
      with self._exec_locals(args):
        return self.eval_node(cnode)
    elif mode == 'eval':
      with self._eval_locals(args):
        return self.eval_node(cnode)
    else:
      fatal(f'Invalid mode: {mode}')

  def flush(self):
    self._flush_generation()
    return self.emitter.flush()

  def emit_code(self, code, **kwargs):
    dcode = textwrap.dedent(code)
    alog.debug(lambda: f'INLINE CODE:\n{dcode}')

    hcode = self._resolve_code(dcode, kwargs)
    self.emitter.emit_code(hcode)

  def emit_report(self, fmt, **kwargs):
    parts = self._get_format_parts(fmt, kwargs)

    self.emitter.emit_report(parts, **kwargs)

  def emit_write(self, fmt, **kwargs):
    parts = self._get_format_parts(fmt, kwargs)

    self.emitter.emit_write(parts, **kwargs)

  def emit_call(self, fname, args, dtype):
    return self.emitter.emit_call(fname, args, dtype)

  def emitter_context(self, **kwargs):
    return self.emitter.context(kwargs)

  def no_hdl(self):
    return self._force_hdl(-1)

  def force_hdl(self):
    return self._force_hdl(1)

  def set_loop_mode(self, mode):
    match mode:
      case 'hdl':
        lmode = _LoopModes.HDL

      case 'unrolled':
        lmode = _LoopModes.UNROLLED

      case _:
        fatal(f'Invalid loop mode: {mode}')

    return self._loop_mode(lmode)

  def generate_name(self, name, shortzero=False):
    return self._revgen.newname(name, shortzero=shortzero)

  def get_frames(self):
    return tuple((f.location.filename, f.location.lineno) for f in self._frames)




import functools


_IS_HDL = '_is_hdl'
_HDL_ARGS = '_hdl_args'


def hdl(func):
  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    return func(*args, **kwargs)

  setattr(func, _IS_HDL, True)
  setattr(wrapper, _IS_HDL, True)

  return wrapper


def hdl_process(**hwargs):
  def decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
      return func(*args, **kwargs)

    setattr(func, _IS_HDL, True)
    setattr(wrapper, _IS_HDL, True)
    setattr(wrapper, _HDL_ARGS, hwargs)

    return wrapper

  return decorator


def is_hdl_function(func):
  return getattr(func, _IS_HDL, False)


def get_hdl_args(func):
  return getattr(func, _HDL_ARGS, None)




import collections
import os
import re
import string
import textwrap

import numpy as np

import py_misc_utils.alog as alog
import py_misc_utils.module_utils as pymu
import py_misc_utils.template_replace as pytr
import py_misc_utils.utils as pyu

from .decorators import *
from .entity import *
from .pyxhdl import *
from .types import *
from .vars import *
from .utils import *
from .wrap import *

from . import xlib as XL


TbData = collections.namedtuple('TbData', 'inputs, outputs, wait, wait_expr, env')


class _TestData:

  def __init__(self, path, eclass):
    self._data = pyu.load_config(path)
    self._conf = self._data.get('conf', dict())
    self._env = self._data.get('env', dict())
    self._loaders = self._conf.get('loaders', dict())

    inp, outp = dict(), dict()
    for pin in eclass.PORTS:
      if pin.is_rd():
        inp[pin.name] = pin
      if pin.is_wr():
        outp[pin.name] = pin

    self._inp, self._outp = inp, outp

  def conf(self, *cpath, defval=None):
    return pyu.dict_rget(self._conf, cpath, defval=defval)

  def setenv(self, name, value):
    self._env[name] = value

  def _load(self, name, value):
    loader = self.conf('loaders', name)
    if loader is not None:
      kind = loader.get('kind')
      if kind == 'numpy':
        return np.array(value, dtype=loader.get('dtype'))
      else:
        alog.warning(f'Unknown loader kind: {kind}')

    return value

  def __iter__(self):
    for data in self._data.get('data', ()):
      inputs, outputs = dict(), dict()
      for k, v in data.items():
        if k in self._inp:
          inputs[k] = self._load(k, v)
        if k in self._outp:
          outputs[k] = self._load(k, v)

      wait = data.get('_wait')
      wait_expr = data.get('_wait_expr')

      yield TbData(inputs=inputs,
                   outputs=outputs,
                   wait=wait,
                   wait_expr=wait_expr,
                   env=self._env)


class _Required:
  pass


_REQUIRED = _Required()

_STD_TOLL = 1e-5

_TB_ARGS = dict(
  input_file=_REQUIRED,
  wait=None,
  clock=[],
  clock_sync=None,
  write_output=False,
  toll=_STD_TOLL,
)


def add_arguments(parser):
  parser.add_argument('--tb_input_file', type=str,
                      help='The testbench input file')
  parser.add_argument('--tb_wait', type=int,
                      help='The amount of wait after setting the inputs, to emit the output')
  parser.add_argument('--tb_clock', type=str, action='append',
                      help='The name and period (format NUM{s,ms,us,ns,ps}) of the clock input (NAME,PERIOD)')
  parser.add_argument('--tb_clock_sync', type=str,
                      help='The edge of the clock to sync the data feed with')
  parser.add_argument('--tb_write_output', action='store_true',
                      help='Whether a write to STDOUT should be generated for every input')
  parser.add_argument('--tb_toll', type=float,
                      help='The tollerance in comparing floating point values')


def _make_args(args):
  tb_args = dict()
  for name, value in _TB_ARGS.items():
    aname = f'tb_{name}'
    avalue = getattr(args, aname, None)
    if avalue is None:
      if value is _REQUIRED:
        fatal(f'Missing command line argument: --{aname}')
      else:
        avalue = value

    tb_args[name] = avalue

  return tb_args


def _get_write_string(eclass, inputs, chunksize=5):
  wparts = []
  for pin in eclass.PORTS:
    dtype = inputs[pin.name].dtype
    if len(dtype.array_shape) > 0:
      for idx in np.ndindex(dtype.array_shape):
        substr = ', '.join(str(x) for x in idx)

        wparts.append(f'{pin.name}[{substr}]={{{pin.name}[{substr}]}}')
    else:
      wparts.append(f'{pin.name}={{{pin.name}}}')

  chunks = [wparts[i: i + chunksize] for i in range(0, len(wparts), chunksize)]
  writes = ['TIME={NOW} ' + ' '.join(chunks[0])]
  for wp in chunks[1: ]:
    writes.append('    ' + ' '.join(wp))

  return '\n'.join(writes)


def _gen_wait(data, wait, clock, clock_sync, eclass, env):
  rwait = data.wait or wait
  if rwait:
    XL.wait_for(rwait)

  if clock_sync == 'rising':
    XL.wait_rising(XL.load(clock))
  elif clock_sync == 'falling':
    XL.wait_falling(XL.load(clock))

  if data.wait_expr is not None:
    wexpr = data.wait_expr.replace(';', '\n')
    XL.xexec(wexpr, **env)


def _repr(v, dtype):
  if isinstance(dtype, Bool):
    return 'true' if v else 'false'

  if isinstance(dtype, Bits):
    fmt = f'{{v:0{dtype.nbits}b}}'
    return fmt.format(v=v)

  if isinstance(dtype, (Sint, Uint)):
    nibbles = (dtype.nbits + 1) // 4
    fmt = f'{{v:0{nibbles}X}}'
    return fmt.format(v=v)

  return v


Clock = collections.namedtuple('Clock', 'name, period')

def _parse_clocks(args):
  clocks = []
  for clk in args['clock']:
    name, period = pyu.resplit(clk, ',')

    clocks.append(Clock(name=name, period=parse_time(period)))

  return clocks


@hdl
def _assign_value(var, value):
  code = []
  if isinstance(value, np.ndarray):
    shape = var.dtype.array_shape
    if tuple(shape) != tuple(value.shape):
      fatal(f'Wrong shape for "{var.ref.name}": {tuple(shape)} vs {tuple(value.shape)}')

    for idx in np.ndindex(shape):
      substr = ', '.join(str(x) for x in idx)
      code.append(f'{var.ref.name}[{substr}] = {value[idx]}')
  else:
    code.append(f'{var.ref.name} = {value}')

  XL.xexec('\n'.join(code))


@hdl
def _values_differ(value, ref_value, toll):
  if isinstance(value.dtype, Float):
    return not XL.float_equal(value, ref_value, toll)
  if isinstance(value.dtype, Real):
    return not XL.real_equal(value, ref_value, toll)

  return value != ref_value


@hdl
def compare_value(var, value, toll=1e-5, debug=None, msg=''):
  cdebug = int(os.getenv('TB_DEBUG', '0')) != 0 if debug is None else debug

  cvar = XL.load(var) if isinstance(var, str) else var

  if isinstance(value, np.ndarray):
    shape = cvar.dtype.array_shape
    if tuple(shape) != tuple(value.shape):
      fatal(f'Wrong shape for "{cvar.value}": {tuple(shape)} vs {tuple(value.shape)}{msg}')

    for idx in np.ndindex(shape):
      tmp = cvar[idx]
      if _values_differ(tmp, value[idx].item(), toll=toll):
        XL.report(f'{{NOW}} Output mismatch: {tmp.value} = {{tmp}} (should be {_repr(value[idx], tmp.dtype)}){msg}')
      elif cdebug:
        XL.report(f'{{NOW}} Output match: {tmp.value} = {{tmp}}{msg}')

      # TODO: Fix the fact we end up with HDL variable write, instead of locals
      # assignment, in case we write the same symbol more than once.
      del tmp
  else:
    if _values_differ(cvar, value, toll=toll):
      XL.report(f'{{NOW}} Output mismatch: {cvar.value} = {{cvar}} (should be {_repr(value, cvar.dtype)}){msg}')
    elif cdebug:
      XL.report(f'{{NOW}} Output match: {cvar.value} = {{cvar}}{msg}')


@hdl
def wait_rising(var):
  # Verilator and GHDL behave differently WRT when the data is available after
  # the edge. This synchronizes at half clock step, which seems to make both
  # happy. If one has strict requirements WRT edge timing, and do not care about
  # compatibility, should use the vanilla XL.wait_rising() API.
  XL.wait_rising(var)
  XL.wait_falling(var)


@hdl
def wait_falling(var):
  # See wait_rising() comment above for rationale of this API.
  XL.wait_falling(var)
  XL.wait_rising(var)


@hdl
def wait_until(clk, test):
  XL.wait_until(test)
  XL.wait_falling(clk)


class TestBench(Entity):

  PORTS = tuple()
  ARGS = dict(args=None)

  def __init__(self, eclass, inputs, args, **kwargs):
    super().__init__(**kwargs)
    self._eclass = eclass
    self._inputs = inputs
    self._tbargs = args
    self._clocks = _parse_clocks(args)

  def _input_data(self):
    _, ext = os.path.splitext(self._tbargs['input_file'])
    if ext == '.py':
      # Inputs comes from a Python file.
      mod = pymu.load_module(self._tbargs['input_file'])

      return mod.tb_iterator(self._eclass, self._inputs, self._tbargs,
                             clocks=self._clocks)

    m = re.match(r'(\w+(\.\w+)*):(\w+)$', self._tbargs['input_file'])
    if m:
      # Inputs comes from a Python module.
      iterfn, = pymu.import_module_names(m.group(1), names=m.group(3))

      return iterfn(self._eclass, self._inputs, self._tbargs,
                    clocks=self._clocks)

    # Inputs comes from a YAML/JSON data file.
    tbdata = _TestData(self._tbargs['input_file'], self._eclass)

    return tbdata

  @hdl_process(kind=INIT_PROCESS)
  def init(self):
    for pin in self._eclass.PORTS:
      XL.assign(pin.name, inputs[pin.name])

  @hdl_process(kind=ROOT_PROCESS)
  def root(self):
    eargs = self._inputs.copy()
    for pin in self._eclass.PORTS:
      eargs[pin.name] = XL.load(pin.name)

    # Instantiate the entity under test.
    ent = self._eclass(**eargs)

  @hdl_process()
  def test(self):
    wait = self._tbargs['wait']
    clock_sync = self._tbargs['clock_sync']
    if clock_sync:
      clock, clock_sync = pyu.resplit(clock_sync, ',')
    else:
      clock = None

    write_string = (_get_write_string(self._eclass, self._inputs)
                    if self._tbargs['write_output'] else None)

    for data in self._input_data():
      for dk, dv in data.inputs.items():
        _assign_value(XL.load(dk), dv)

      _gen_wait(data, wait, clock, clock_sync, self._eclass, data.env)

      if write_string is not None and (data.inputs or data.outputs):
        for wstr in write_string.split('\n'):
          XL.write(wstr)

      for dk, dv in data.outputs.items():
        compare_value(dk, dv, toll=args['toll'])

    XL.finish()

  _CLOCK_FN = """
    @hdl_process()
    def $clock_fn(self):
      $clk_name = not $clk_name
      XL.wait_for($period / 2)
  """

  def enum_processes(self):
    for pfn in super().enum_processes():
      yield pfn

    env = globals().copy()
    for clk in self._clocks:
      alog.debug(f'Generating clock "{clk.name}" with period {clk.period}')

      clock_fn = f'clock_{clk.name}'
      scode = pytr.template_replace(textwrap.dedent(self._CLOCK_FN),
                                    vals=dict(clock_fn=clock_fn,
                                              clk_name=clk.name,
                                              period=clk.period,
                                              ))

      pfn, = pyu.compile(scode, clock_fn, env=env)

      # The inspect API is not able to find the source code information required
      # by the pyxhdl module when using exec(), so we need to provide them here.
      set_function_info(pfn.__wrapped__, __file__, 1, scode)

      yield pfn


def generate(codegen, args, eclass, inputs):
  tb_args = dict(args=_make_args(args), eclass=eclass, inputs=inputs)

  codegen.generate_entity(TestBench, tb_args)




import collections

import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .decorators import *
from .port import *
from .types import *
from .vars import *
from .utils import *


ArgPort = collections.namedtuple('ArgPort', 'arg, port')

POSEDGE = 'POSEDGE'
NEGEDGE = 'NEGEDGE'
LEVEL = 'LEVEL'

Sens = collections.namedtuple('Sens', 'trigger', defaults=[LEVEL])


class _CoreEntity:

  def __init_subclass__(cls):
    cls.PORTS = cls._get_ports(cls.PORTS)

  @classmethod
  def _get_ports(cls, cports):
    return Port.parse_list(cports)


class Entity(_CoreEntity):

  PORTS = tuple()
  ARGS = dict()
  NAME = None
  LIBNAME = None

  def __init__(self, **kwargs):
    super().__init__()
    self.args = dict()
    self.kwargs = dict()

    for pin in self.PORTS:
      arg = kwargs.get(pin.name)
      if arg is None:
        fatal(f'Missing argument "{pin.name}" for Entity "{pyiu.cname(self)}"')

      if pin.is_ifc():
        arg = pin.ifc_view(arg)

      self.args[pin.name] = ArgPort(arg, pin)

    for arg_name, arg in self.ARGS.items():
      self.kwargs[arg_name] = kwargs.get(arg_name, arg)

  def __repr__(self):
    gargs = self.kwargs.copy()
    gargs.update(self.args)

    rstr = pyu.stri(gargs)

    return pyiu.cname(self) + '(' + rstr[1: -1] + ')'

  def expanded_ports(self):
    xports = []
    for name, aport in self.args.items():
      if aport.port.is_ifc():
        for pin, arg in aport.port.ifc_expand(aport.arg):
          xports.append(ArgPort(arg, pin))
      else:
        xports.append(aport)

    return tuple(xports)

  def enum_processes(self):
    for name, func in self.__class__.__dict__.items():
      if is_hdl_function(func):
        if needs_self(func):
          func = getattr(self, name)

        yield func




# Cannot depend of local modules!
import numpy as np


class Wrapper:
  pass


class _NumpyWrapper(Wrapper):

  def __init__(self, value):
    super().__init__()
    self.value = value

  def __eq__(self, other):
    return self.value == other.value

  def __str__(self):
    return str(self.value.tolist())

  def __hash__(self):
    return hash((str(self.value.dtype), tuple(self.value.shape), self.value.tobytes()))



def unwrap(value):
  if isinstance(value, Wrapper):
    return value.value

  # Numpy scalars expose themselves as Python 'float, int' but are rejected by compile()
  # as they show up as 'np.float*, np.int*'.
  if isinstance(value, np.number):
    return value.item()

  return value


def wrap(value):
  if isinstance(value, np.ndarray):
    return _NumpyWrapper(value)

  return value




from .decorators import *
from .emitter import *
from .pyxhdl import *
from .types import *
from .utils import *
from .vars import *
from . import xlib as XL


@hdl
def bit_swap(src):
  dest = mkwire(src.dtype, name=XL.generate_name('bit_swap'))
  for i in range(dest.dtype.nbits):
    dest[i] = src[dest.dtype.nbits - 1 - i]

  return dest




import re

import py_misc_utils.core_utils as pycu


class _Instance:

  def __init__(self, name, params, args):
    self.name = name
    self.params = params
    self.args = args
    self._hash = pycu.genhash((name, params, args))

  def __hash__(self):
    return self._hash

  def __eq__(self, other):
    return (self.name == other.name and self.params == other.params and
            self.args == other.args)


class Instanciator:

  def __init__(self, param_key=None, revbase=None):
    self._param_key = param_key
    self._instdb = dict()
    self._instances = []
    self._revgen = pycu.RevGen(revbase=revbase if revbase is not None else 1)

  def _cname(self, name):
    cname = re.sub(r'\W+', '_', name)

    return cname

  def getid(self, name, kwargs, force_new=False):
    params = kwargs.pop(self._param_key, dict()) if self._param_key else dict()
    inst = _Instance(name, params, kwargs)
    if force_new:
      cname = self._cname(name)
      iid = self._revgen.newname(cname)

      self._instances.append((iid, inst))
    else:
      iid = self._instdb.get(inst)
      if iid is None:
        cname = self._cname(name)
        iid = self._revgen.newname(cname)

        self._instdb[inst] = iid
        self._instances.append((iid, inst))

    return iid

  def __iter__(self):
    yield from self._instances




import re

import py_misc_utils.core_utils as pycu
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.module_utils as pymu
import py_misc_utils.utils as pyu

from .types import *
from .vars import *
from .utils import *


class Port:

  __slots__ = ('name', 'idir', 'type', 'default')

  IN = 'IN'
  OUT = 'OUT'
  INOUT = 'INOUT'
  IFC = 'IFC'

  def __init__(self, name, idir, type=None, default=None):
    self.name = name
    self.idir = idir
    self.type = type
    self.default = default

  def __repr__(self):
    rfmt = pyu.repr_fmt(self, 'name=,idir=,type,default')

    return f'{pyiu.cname(self)}({rfmt})'

  def __hash__(self):
    return hash((self.name, self.idir, self.type, self.default))

  def __eq__(self, other):
    return (self.name == other.name and self.idir == other.idir and
            self.type == other.type and self.default == other.default)

  def is_ro(self):
    return self.idir == self.IN

  def is_wo(self):
    return self.idir == self.OUT

  def is_rd(self):
    return self.idir in (self.IN, self.INOUT)

  def is_wr(self):
    return self.idir in (self.OUT, self.INOUT)

  def is_rw(self):
    return self.idir == self.INOUT

  def is_ifc(self):
    return self.idir == self.IFC

  def ifc_split(self):
    ifc_class, ifc_port = pycu.separate(self.type, '.', reverse=True)

    return ifc_class, ifc_port

  def ifc_view(self, value):
    ifc_class, ifc_port = self.ifc_split()

    return value.origin.create_port_view(self.name, ifc_port)

  def ifc_expand(self, value):
    ifc_class, ifc_port = self.ifc_split()

    return value.origin.expand_port(self.name, ifc_port)

  def ifc_field_name(self, fname):
    return subname(self.name, fname)

  def verify_arg(self, arg):
    if self.type is not None:
      if self.is_ifc():
        ifc_class, ifc_port = self.ifc_split()

        pcls, = pymu.import_module_names(ifc_class)

        if not isinstance(arg.origin, pcls):
          fatal(f'Invalid argument of type {pyiu.cname(arg.origin)} ' \
                'when {ifc_class} is required')
      else:
        tmatch = TypeMatcher.parse(self.type)
        tmatch.check_value(arg, msg=f' for entity port "{self.name}"')

  @classmethod
  def parse(cls, pdecl):
    m = re.match(r'(=|\+|\*)?(\w+)(:([^\s]*))?(\s*=\s*(\w+))?$', pdecl)
    if not m:
      fatal(f'Unrecognized port format: {pdecl}')

    match m.group(1):
      case '=':
        idir = cls.OUT
      case '+':
        idir = cls.INOUT
      case '*':
        idir = cls.IFC
      case _:
        idir = cls.IN

    type_default = dtype_from_string(m.group(6)) if m.group(6) else None

    return cls(m.group(2), idir, type=m.group(4), default=type_default)

  @classmethod
  def parse_list(cls, cports):
    if isinstance(cports, (list, tuple)):
      if not all(isinstance(p, cls) for p in cports):
        fatal(f'Unrecognized port value: {cports}')

      return tuple(cports)
    elif isinstance(cports, str):
      ports = []
      for pdecl in pyu.resplit(cports, ','):
        ports.append(cls.parse(pdecl))

      return tuple(ports)
    else:
      fatal(f'Unrecognized ports value: {cports}')


def make_port_ref(pin):
  match pin.idir:
    case Port.IN:
      mode = Ref.RO

    case Port.OUT:
      mode = Ref.WO

    case Port.INOUT:
      mode = Ref.RW

    case _:
      fatal(f'Unrecognized Port direction: {pin.idir}')

  return Ref(pin.name, mode=mode, vspec=VSpec(port=pin), vname=pin.name)

